#include "StdAfx.h"
#include "resource.h"
#include "ConfigData.h"
#include "Persist.h"
#include "PromptDlg.h"

static LPCTSTR ConfigFileName = TEXT("LiteEdit.ini");

LanguagePtr LanguageMgr::FindLanguage(const CString& ext, LanguagePtr skipLang/* = NULL*/) {
	for (size_t i = 0; i < mLangs.size(); ++i)
		if (mLangs[i] != skipLang && mLangs[i]->HasExtension(ext))
			return mLangs[i];

	return NULL;
}

void LanguageMgr::Load() {
	mLangs.clear();

	//declare these outside the try/catch so the file name will be available for
	//the error message
	StringVector langFiles;
	FindAllFiles(GetLanguagePath(TEXT("*")), langFiles, true);
	size_t i = 0;

	try {
		for (i = 0; i < langFiles.size(); ++i) {
			Persist p(langFiles[i], true);
			LanguagePtr lang = new Language();
			lang->DoPersist(p);
			mLangs.push_back(lang);
		}
	} catch (Exception& e) {
		if (e.mErrorCode == ErrNewerFileVersion) {
			CString msg;
			msg.Format(TEXT("The version of file '%s' is newer than the version of %s. %s could not load one or more language files."), langFiles[i], PROGRAM_NAME, PROGRAM_NAME);
			DisplayError(msg);
		} else
			e.DisplayDefaultError();
	}
}

void LanguageMgr::Save() {
	for (size_t i = 0; i < mLangs.size(); ++i) {
		try {
			Persist p(GetLanguagePath(mLangs[i]->GetName()), false);
			mLangs[i]->DoPersist(p);
		//it's important that exceptions are cought because the .lang files have been
		//deleted, so the user would lose .lang files if an exception excaped before
		//all the .lang files have been written to disk
		} catch (CException* e) {
			e->ReportError(MB_OK|MB_ICONERROR);
		}
	}
}

CString LanguageMgr::BuildFilter() const {
	CString filter = AllFilesFilter;
	bool hasTextFiles = false;

	for (size_t i1 = 0; i1 < mLangs.size(); ++i1) {
		const StringVector& exts = mLangs[i1]->m_extensions;
		if (exts.empty())
			continue;

		CString filterExt;
		CString filterDesc = mLangs[i1]->GetName() + TEXT(" Files (");
		for (size_t i2 = 0; i2 < exts.size(); ++i2) {
			if (StrCmpI(exts[i2], TEXT("txt")) == 0)
				hasTextFiles = true;

			filterExt += TEXT("*.") + exts[i2];
			filterDesc += exts[i2];

			if (i2 < exts.size() - 1) {
				filterExt += ';';
				filterDesc += TEXT("; ");
			}
		}

		filterDesc += ')';
		filter += filterDesc + '|' + filterExt + '|';
	}

	if (!hasTextFiles)
		filter += TEXT("Text Files (txt)|*.txt|");

	filter += '|';
	return filter;
}

void RecentFileList::ValidateSize() {
	if ((long)mFiles.size() > mMax) {
		long numToErase = mFiles.size() - mMax;
		mFiles.erase(mFiles.end() - numToErase, mFiles.end());
	}
}

void RecentFileList::PergeNonExistantFiles() {
	size_t i = 0;
	while (i < mFiles.size())
		//don't check if the file exists if it is not on this machine since this
		//function gets called on start up and shouldn't be slow if the user's
		//network connection is slow
		if (IsLocalFile(mFiles[i]) && !FileExists(mFiles[i]))
			mFiles.erase(mFiles.begin() + i);
		else
			++i;
}

void RecentFileList::SetMax(long val) {
	mMax = Max(val, 0L);
	ValidateSize();
}

void RecentFileList::Add(const CString& fileName) {
	ASSERT(AbsolutePathExists(fileName));

	Remove(fileName);
	mFiles.insert(mFiles.begin(), fileName);
	ValidateSize();
}

void RecentFileList::Remove(const CString& fileName) {
	long i = mFiles.FindString(fileName, false);
	if (i >= 0)
		mFiles.erase(mFiles.begin() + i);
}

HMENU RecentFileList::CreateMenu(CommandIdCreator& idCreator) const {
	CMenu menu;
	VERIFY(menu.CreatePopupMenu());

	for (size_t i = 0; i < mFiles.size(); ++i) {
		CString file = mFiles[i];
		if (i < 10)
			file.Format(TEXT("&%d %s"), i, file);
		else if (i <= 'Z' - 'A' + 10)
			file.Format(TEXT("&%c %s"), i - 10 + 'A', file);

		VERIFY(menu.AppendMenu(MF_STRING|MF_ENABLED, idCreator.CreateId(), file));
	}

	return menu.Detach();
}

void RecentFileList::DoPersist(Persist& p) {
	p.PersistNumber(TEXT("MaxRecentFiles"), mMax);
	p.PersistValVector(TEXT("RecentFiles"), mFiles);

	if (p.IsLoading())
		PergeNonExistantFiles();
}

//Pass true for displayStringOnly for %Prompt% to be replaced with [Prompt] instead
//of actually prompting the user for input.
CString GetCommandString(const CString& curFileName, const CString& format, bool displayStringOnly/* = false*/) {
	CString result = ExpandEnvironmentStrings(format);

	TCHAR curDir[MAX_PATH];
	VERIFY(GetCurrentDirectory(MAX_PATH, curDir));

	CString filePath = curFileName;
	CString fileDir, fileName, fileExt;
	BreakPath(filePath, fileDir, fileName, fileExt);

	//do replacment for current directoryu and path macros
	ReplaceSubstring(result, CurDirMacro, curDir, ffIgnoreCase);
	ReplaceSubstring(result, FilePathMacro, filePath, ffIgnoreCase);
	ReplaceSubstring(result, FileDirMacro, fileDir, ffIgnoreCase);
	ReplaceSubstring(result, FileNameMacro, fileName, ffIgnoreCase);
	ReplaceSubstring(result, FileExtMacro, fileExt, ffIgnoreCase);

	//do replacement for prompt macro
	long pos = 0;
	while (pos < result.GetLength()) {
		pos = FindSubstring(result, PromptMacroStart, ffIgnoreCase, pos);
		if (pos < 0)
			break;
		
		long promptStart = pos;
		long promptStrStart = promptStart + PromptMacroStart.GetLength();
		if (promptStrStart >= result.GetLength())
			break;

		pos = FindSubstring(result, TEXT("%"), ffIgnoreCase, promptStrStart);
		if (pos < 0)
			break;
		
		CString promptStr = result.Mid(promptStrStart, pos - promptStrStart);

		CString replaceStr;
		if (displayStringOnly)
			replaceStr.Format(TEXT("<%s>"), promptStr);
		else {
			PromptDlg dlg(promptStr);
			VERIFY(dlg.DoModal() == IDOK);
			replaceStr = dlg.mText;
		}

		result.Delete(promptStart, pos - promptStart + 1);
		result.Insert(promptStart, replaceStr);
		pos = promptStart + replaceStr.GetLength();
	}

	return result;
}

void Command::DoPersist(Persist& p) {
	p.PersistString(TEXT("MenuItemName"), mMenuItemName);
	p.PersistString(TEXT("Command"), mCommand);
	p.PersistString(TEXT("Arguments"), mArguments);
	p.PersistString(TEXT("InitDir"), mInitDir);
	p.PersistBool(TEXT("SaveFileBeforeExecuting"), mSaveFile);
	p.PersistValVector(TEXT("Extensions"), mExts);

	mAccel.DoPersist(p);
	if (p.IsLoading())
		mAccel.cmd = 0;//I only need to persist the hotkey
}

void Commands::AddCommandsToMenu(HMENU hMenu, CommandIdCreator& idCreator) {
	CMenu menu;
	VERIFY(menu.Attach(hMenu));

	mAccelTable.Clear();

	for (size_t i = 0; i < size(); ++i) {
		UINT id = idCreator.CreateId();
		Item(i).mAccel.cmd = id;
		VERIFY(menu.InsertMenu(i, MF_STRING|MF_BYPOSITION, id, Item(i).mMenuItemName));
		mAccelTable.AddAccel(Item(i).mAccel);
	}

	mAccelTable.DecorateMenu(menu);

	menu.Detach();
}

Command* Commands::FindCommand(UINT cmd) {
	for (size_t i = 0; i < size(); ++i)
		if (Item(i).mAccel.cmd == cmd)
			return &Item(i);

	return NULL;
}

void Commands::DoPersist(Persist& p) {
	p.PersistClassVector(TEXT("Commands"), *this);
}

void FindReplaceData::DoPersist(Persist& p) {
	p.PersistBool(TEXT("AllowEscapedChars"), escapedChars);
	p.PersistBool(TEXT("CloseDialogOnFind"), closeOnFind);
}

void ColorScheme::DoPersist(Persist& p) {
	p.PersistString(TEXT("Name"), mName);
	p.PersistEnum(TEXT("SelColor"), mSelColor);
	TokenColors::DoPersist(p);
}

ColorSchemes& ColorSchemes::operator=(const ColorSchemes& rhs) {
	clear();
	for (size_t i = 0; i < rhs.size(); ++i)
		push_back(new ColorScheme(*rhs[i]));

	mCurScheme = rhs.mCurScheme;
	return *this;
}

void ColorSchemes::DoPersist(Persist& p) {
	p.PersistNumber(TEXT("CurrentColorScheme"), mCurScheme);
	p.PersistPtrVector(TEXT("ColorScheme"), *this);
}

CString GetTimeDateString(const CString& format) {
	CString string = format;

	ReplaceSubstring(string, DateMacro, GetSystemDateString(false), ffIgnoreCase);
	ReplaceSubstring(string, LongDateMacro, GetSystemDateString(true), ffIgnoreCase);
	ReplaceSubstring(string, TimeMacro, GetSystemTimeString(false), ffIgnoreCase);
	ReplaceSubstring(string, LongTimeMacro, GetSystemTimeString(true), ffIgnoreCase);

	return string;
}

//initialized default values for all config data
ConfigData::ConfigData() {
	mDefExt = TEXT("txt");
	mTimeDateFormat = CString(DateMacro) + TEXT(" ") + TimeMacro;
	mRemoveTrailingWhitespaceOnSave = true;
	mNewLineType = nlCRLF;
	mAutoIndent = true;
	mTabWidth = 4;
	mMaximized = false;
	mOpenFilterIndex = 0;
	mShowLineNumbers = false;

	LOGFONT lf;
	MemClear(lf);
	StrCpy(lf.lfFaceName, TEXT("Courier New"));
	lf.lfHeight = -12;
	lf.lfWeight = FW_NORMAL;
	VERIFY(mFont.CreateFontIndirect(&lf));
	//mFont.Attach((HFONT)GetStockObject(ANSI_FIXED_FONT));

	for (int i = 0; i < MAX_CUSTOM_COLORS; ++i)
		mCustomColors[i] = RGB(0xFF, 0xFF, 0xFF);

	mAccelTable.AddAccel(Accel('N', kmCtrl, ID_FILE_NEW));
	mAccelTable.AddAccel(Accel('O', kmCtrl, ID_FILE_OPEN));
	mAccelTable.AddAccel(Accel('S', kmCtrl, ID_FILE_SAVE));

	mAccelTable.AddAccel(Accel('Z', kmCtrl, ID_EDIT_UNDO));
	mAccelTable.AddAccel(Accel('Y', kmCtrl, ID_EDIT_REDO));
	mAccelTable.AddAccel(Accel('X', kmCtrl, ID_EDIT_CUT));
	mAccelTable.AddAccel(Accel('C', kmCtrl, ID_EDIT_COPY));
	mAccelTable.AddAccel(Accel('V', kmCtrl, ID_EDIT_PASTE));
	mAccelTable.AddAccel(Accel('A', kmCtrl, ID_EDIT_SELECTALL));
	mAccelTable.AddAccel(Accel('W', kmCtrl, ID_EDIT_ADVANCED_SWITCHCLIPBOARD));
	mAccelTable.AddAccel(Accel('D', kmCtrl|kmShift, ID_EDIT_ADVANCED_INSERTDATETIME));
	mAccelTable.AddAccel(Accel('L', kmCtrl|kmShift, ID_EDIT_ADVANCED_LOWERCASE));
	mAccelTable.AddAccel(Accel('U', kmCtrl|kmShift, ID_EDIT_ADVANCED_UPPERCASE));
	mAccelTable.AddAccel(Accel('T', kmCtrl|kmShift, ID_EDIT_ADVANCED_TITLECASE));

	mAccelTable.AddAccel(Accel('F', kmCtrl, ID_SEARCH_FIND));
	mAccelTable.AddAccel(Accel(VK_F3, kmNone, ID_SEARCH_FINDNEXT));
	mAccelTable.AddAccel(Accel(VK_F3, kmShift, ID_SEARCH_FINDPREVIOUS));
	mAccelTable.AddAccel(Accel('H', kmCtrl, ID_SEARCH_REPLACE));
	mAccelTable.AddAccel(Accel('G', kmCtrl, ID_SEARCH_GOTO));

	mAccelTable.AddAccel(Accel(VK_F2, kmCtrl, ID_BOOKMARKS_TOGGLE));
	mAccelTable.AddAccel(Accel(VK_F2, kmNone, ID_BOOKMARKS_NEXT));
	mAccelTable.AddAccel(Accel(VK_F2, kmShift, ID_BOOKMARKS_PREVIOUS));

	mAccelTable.AddAccel(Accel(VK_F1, kmNone, ID_HELP_CONTENTS));

	LPCTSTR commandFileName = TEXT("cmd.exe");
	if (!SearchForFile(CString(commandFileName)))
		commandFileName = TEXT("Command.com");
	mCommands.push_back(Command(commandFileName, TEXT("&Console Window"), EmptyStr, FileDirMacro, false, 'C', kmShift|kmCtrl));
	mCommands.push_back(Command(TEXT("explorer.exe"), TEXT("&Windows Explorer"), TEXT("/e, ") + Quote(FileDirMacro), EmptyStr, false, 'E', kmCtrl));
	mCommands.push_back(Command(GetInternetExplorerPath(), TEXT("&Internet Explorer"), Quote(FilePathMacro), FileDirMacro, true, 'I', kmCtrl));
	mCommands.push_back(Command(GetFileNameWithExt(GetModulePath()), TEXT("&Lite Edit"), EmptyStr, EmptyStr, false, 'E', kmShift|kmCtrl));

	Command exeCurFile(TEXT("%FileDir%%FileName%.exe"), TEXT("E&xecute Current File"), EmptyStr, FileDirMacro, true, VK_F5, kmNone);
	exeCurFile.mExts.push_back(TEXT("c"));
	exeCurFile.mExts.push_back(TEXT("cpp"));
	exeCurFile.mExts.push_back(TEXT("pas"));
	mCommands.push_back(exeCurFile);
}

void ConfigData::DoPersist(bool load) {
	CString fileName = GetModuleDir() + ConfigFileName;
	bool useSystemSelectionColor = false;

	try {
		if (!load || FileExists(fileName)) {
			Persist p(fileName, load);

			p.PersistString(TEXT("DefaultExtention"), mDefExt);
			p.PersistString(TEXT("TimeDateFormat"), mTimeDateFormat);
			p.PersistBool(TEXT("RemoveTrailingWhitespaceOnSave"), mRemoveTrailingWhitespaceOnSave);
			p.PersistEnum(TEXT("NewLineType"), mNewLineType);
			p.PersistBool(TEXT("AutoIndent"), mAutoIndent);
			p.PersistNumber(TEXT("TabWidth"), mTabWidth);
			p.PersistBool(TEXT("Maximized"), mMaximized);
			p.PersistNumber(TEXT("OpenFilterIndex"), mOpenFilterIndex);

			if (p.IsVersBefore(PersistVers3))
				p.PersistBool(TEXT("UseSystemSelectionColor"), useSystemSelectionColor);

			p.PersistNewLine();

			p.PersistVal(TEXT("WinPos"), mWinPos);

			p.PersistNewLine();

			mRecentFileList.DoPersist(p);

			p.PersistNewLine();

			ASSERT(sizeof(HFONT) == sizeof(mFont.m_hObject));//make sure reference cast is safe
			p.PersistVal(TEXT("Font"), (HFONT&)mFont.m_hObject);

			p.PersistNewLine();

			if (p.IsVersBefore(PersistVers3)) {
				//just a TokenColors used to be persisted
				ColorSchemePtr customColorScheme = new ColorScheme(TEXT("Custom Color Scheme"), useSystemSelectionColor ? scSystem : scInvert);
				customColorScheme->TokenColors::DoPersist(p);
				mColorSchemes.push_back(customColorScheme);
				mColorSchemes.SetCurScheme(mColorSchemes.size() - 1);
			} else
				mColorSchemes.DoPersist(p);

			p.PersistNewLine();

			p.PersistStaticNumberArray(TEXT("CustomColors"), mCustomColors, MAX_CUSTOM_COLORS);

			p.PersistNewLine();

			mAccelTable.DoPersist(p);

			p.PersistNewLine();

			mCommands.DoPersist(p);

			p.PersistNewLine();

			mFindReplaceData.DoPersist(p);

			if (p.IsVersAtLeast(PersistVers2)) {
				if (p.IsVersAtLeast(PersistVers4))
					p.PersistNewLine();

				p.PersistBool("ShowLineNumbers", mShowLineNumbers);
			}

			if (load) {
				if (p.IsVersBefore(CurPersistVer))
					AddConfigDataForOldVersions(p.GetVersion());
			}
		} else if (load) {//loading but not ini file exists
			ASSERT(!FileExists(fileName));

			//pass first persist version so that all new config data will be added
			AddConfigDataForOldVersions(PersistVers1);
		}
	} catch (Exception& e) {
		if (e.mErrorCode == ErrNewerFileVersion) {
			CString msg;
			msg.Format(TEXT("The version of file '%s' is newer than the version of %s. %s will be overwritten. You should back up %s now if you don't want to lose your configuration data."), fileName, PROGRAM_NAME, ConfigFileName, ConfigFileName);
			DisplayError(msg);
		} else
			e.DisplayDefaultError();
	} catch (CException* e) {
		e->ReportError(MB_OK|MB_ICONERROR);
	}

	//mLangMgr gets saved by the LanguagesPage
	if (load)
		mLangMgr.Load();
}

void ConfigData::SetEdit(EditCtrl& edit) {
	edit.SetFont(mFont);
	edit.SetAutoIndent(mAutoIndent);
	edit.SetTabWidth(mTabWidth);
	edit.SetShowLineNumbers(mShowLineNumbers);

	ColorSchemePtr curCS = mColorSchemes.GetCurColorScheme();
	edit.SetUseSystemSelectionColors(curCS == NULL || mColorSchemes.GetCurColorScheme()->GetSelColor() == scSystem);
	edit.SetTokenColors(curCS.GetPtr());
}

//Edit>>Switch Clipboard command.
void ConfigData::SwitchClipboard(HWND hwnd) {
	CString clipboard;
	bool succ = GetClipboardString(clipboard, hwnd, false);
	if (succ) {
		succ = SetClipboardString(mTempClip, hwnd, false);
		if (succ)
			mTempClip = clipboard;
	}

	ATLASSERT(succ);
}

void ConfigData::Execute(const Command& cmd, const CString& fileName, HWND hwnd) const {
	CString progName = GetCommandString(fileName, cmd.mCommand);
	SearchForFile(progName);
	progName = Quote(progName);
	CString commandLine = GetCommandString(fileName, cmd.mArguments);
	commandLine = progName + " " + commandLine;
	LPTSTR commandLineBuff = commandLine.GetBuffer(commandLine.GetLength());
	CString initDirString = GetCommandString(fileName, cmd.mInitDir);
	LPCTSTR initDir = initDirString.IsEmpty() ? NULL : (LPCTSTR)initDirString;
	DWORD flags = 0;
	STARTUPINFO startUpInfo;
	PROCESS_INFORMATION procInfo;
	MemClear(startUpInfo);
	startUpInfo.cb = sizeof(startUpInfo);

	BOOL succ = CreateProcess(NULL, commandLineBuff, NULL, NULL, FALSE, flags, NULL, initDir, &startUpInfo, &procInfo);
	commandLine.ReleaseBuffer();

	if (!succ) {
		DWORD err = GetLastError();
		CString commandString = GetCommandString(fileName, cmd.mCommand);
		CString msg;
		msg.Format("Error executing command '%s':\n\n%s", commandString, GetErrorMessage(err));
		MessageBox(hwnd, msg, ErrorStr, MB_OK|MB_ICONERROR);
	}
}

void ConfigData::ExecuteHelp(const CString& topic, HWND hwnd) const {
	Command helpCommand(GetInternetExplorerPath());
	CString arg = GetModuleDir() + TEXT("LiteEditHelp.html");
	AppendString(arg, topic, TEXT("#"));
	arg = TEXT("\"") + arg + TEXT("\"");
	helpCommand.mArguments = arg;
	Execute(helpCommand, EmptyStr, hwnd);
}

void ConfigData::AddConfigDataForOldVersions(long version) {
	if (version < PersistVers2) {
		mAccelTable.AddAccel(Accel('G', kmCtrl|kmShift, ID_EDIT_ADVANCED_TOGGLECASE));

		CString charMapFileName = TEXT("%SystemRoot%\\System32\\charmap.exe");
		if (FileExists(ExpandEnvironmentStrings(charMapFileName)))
			mCommands.push_back(Command(charMapFileName, TEXT("Character &Map"), EmptyStr, EmptyStr, false, 'M', kmCtrl));
	}

	if (version < PersistVers3) {
		ColorSchemePtr cs;

		//pac man colors
		cs = new ColorScheme(TEXT("Pac Man"), scInvert);
		cs->Item(tkPlainText) = RGB(255, 255, 0);
		cs->Item(tkComment) = RGB(0, 255, 255);
		cs->Item(tkKeyword) = RGB(255, 255, 255);
		cs->Item(tkSymbol) = RGB(255, 255, 255);
		cs->Item(tkString) = RGB(64, 255, 64);
		cs->Item(tkBackground) = RGB(0, 0, 172);
		mColorSchemes.push_back(cs);

		//windows looking default colors
		cs = new ColorScheme(TEXT("T-Shirt"), scSystem);
		cs->Item(tkPlainText) = RGB(0x0, 0x0, 0x0);
		cs->Item(tkComment) = RGB(0x0, 0x80, 0x0);
		cs->Item(tkKeyword) = RGB(0x0, 0x0, 0xFF);
		cs->Item(tkSymbol) = RGB(0xC0, 0x00, 0xC0);
		cs->Item(tkString) = RGB(0x60, 0x60, 0xC0);
		cs->Item(tkBackground) = RGB(0xFF, 0xFF, 0xFF);
		mColorSchemes.push_back(cs);

		//space
		cs = new ColorScheme(TEXT("Outer Space"), scInvert);
		cs->Item(tkPlainText) = RGB(255, 255, 255);
		cs->Item(tkComment) = RGB(176, 176, 255);
		cs->Item(tkKeyword) = RGB(128, 255, 128);
		cs->Item(tkSymbol) = RGB(255, 128, 255);
		cs->Item(tkString) = RGB(255, 255, 0);
		cs->Item(tkBackground) = RGB(0, 0, 0);
		mColorSchemes.push_back(cs);

		//Nature
		cs = new ColorScheme(TEXT("Nature"), scInvert);
		cs->Item(tkPlainText) = RGB(0, 32, 0);
		cs->Item(tkComment) = RGB(255, 0, 0);
		cs->Item(tkKeyword) = RGB(0, 0, 255);
		cs->Item(tkSymbol) = RGB(255, 0, 255);
		cs->Item(tkString) = RGB(64, 32, 0);
		cs->Item(tkBackground) = RGB(212, 255, 212);
		mColorSchemes.push_back(cs);
	}
}

ConfigData gConfigData;