#ifndef ConfigDataH
#define ConfigDataH

#include "EditCtrl.h"

static LPCTSTR LangExt = TEXT(".lang");

inline CString GetLanguagePath(const CString& languageName) {
	return GetModuleDir() + languageName + LangExt;
}

typedef vector<LanguagePtr> Languages;

class LanguageMgr {
private:
public:
	Languages mLangs;

	LanguagePtr FindLanguage(const CString& ext, LanguagePtr skipLang = NULL);
	void Load();
	void Save();
	CString BuildFilter() const;
};

class RecentFileList {
private:
	long mMax;
	StringVector mFiles;

	void ValidateSize();
	void PergeNonExistantFiles();

public:
	RecentFileList() : mMax(15) {}

	long GetMax() const {return mMax;}
	void SetMax(long val);

	void Add(const CString& fileName);
	void Remove(const CString& fileName);
	const StringVector& GetItems() const {return mFiles;}
	HMENU CreateMenu(CommandIdCreator& idCreator) const;
	void DoPersist(Persist& p);
};

static LPCTSTR FilePathMacro = TEXT("%FilePath%");
static LPCTSTR FileDirMacro = TEXT("%FileDir%");
static LPCTSTR FileNameMacro = TEXT("%FileName%");
static LPCTSTR FileExtMacro = TEXT("%FileExt%");
static LPCTSTR CurDirMacro = TEXT("%CurDir%");
static CString PromptMacroStart = TEXT("%Prompt:");

CString GetCommandString(const CString& curFileName, const CString& format, bool displayStringOnly = false);

class Command {
public:
	CString mMenuItemName;
	CString mCommand;
	CString mArguments;
	CString mInitDir;
	bool mSaveFile;
	StringVector mExts;
	Accel mAccel;

	Command(const CString& command = EmptyStr, const CString& menuItemName = TEXT("New &Tool"), const CString& arguments = Quote(FilePathMacro), const CString& initDir = FileDirMacro, bool saveFile = true, VirtualKey key = 0, KeyModifiers mods = kmNone) :
		mMenuItemName(menuItemName), mCommand(command), mArguments(arguments), mInitDir(initDir), mSaveFile(saveFile), mAccel(key, mods, 0) {}

	bool HasExtension(const CString& ext) const {return mExts.HasString(ext, false);}
	void DoPersist(Persist& p);
};

class Commands : public Vector<Command> {
public:
	AccelTable mAccelTable;

	void AddCommandsToMenu(HMENU hMenu, CommandIdCreator& idCreator);
	Command* FindCommand(UINT cmd);
	void DoPersist(Persist& p);
};

struct FindReplaceData {
	//the following are never read by the dialog and so don't need to be initialized
	bool findNext;
	bool replaceCurrent;
	bool replaceAll;
	CString findStr;
	CString replaceStr;

	StringVector findStrings;
	StringVector replaceStrings;
	FindFlags flags;
	bool selectionOnly;
	bool closeOnFind;//persisted
	bool escapedChars;

	FindReplaceData() :
		flags(ffIgnoreCase), selectionOnly(false), closeOnFind(true), escapedChars(true) {}

	void DoPersist(Persist& p);
};

enum SelColorKinds {scInvert, scSystem};

class ColorScheme : public TokenColors {
private:
	CString mName;
	SelColorKinds mSelColor;

public:
	ColorScheme(const CString& name = EmptyStr, SelColorKinds sck = scInvert) : mName(name), mSelColor(sck) {}

	const CString& GetName() const {return mName;}
	void SetName(const CString& name) {mName = name;}
	SelColorKinds GetSelColor() const {return mSelColor;}
	void SetSelColor(SelColorKinds scKind) {mSelColor = scKind;}

	virtual void DoPersist(Persist& p);
};
PTR_DEF(ColorScheme);

class ColorSchemes : public Vector<ColorSchemePtr> {
private:
	long mCurScheme;

public:
	ColorSchemes() : mCurScheme(0) {}
	ColorSchemes(const ColorSchemes& rhs) {*this = rhs;}

	ColorSchemes& operator=(const ColorSchemes& rhs);

	ColorSchemePtr GetCurColorScheme() const {
		return mCurScheme >= 0 ? Item(mCurScheme) : NULL;
	}

	long GetCurScheme() const {return mCurScheme;}

	void SetCurScheme(long val) {
		ATLASSERT(-1 <= val && val < (long)size());
		mCurScheme = val;
	}

	void DoPersist(Persist& p);	
};

static LPCTSTR DateMacro = TEXT("%Date%");
static LPCTSTR LongDateMacro = TEXT("%LongDate%");
static LPCTSTR TimeMacro = TEXT("%Time%");
static LPCTSTR LongTimeMacro = TEXT("%LongTime%");

CString GetTimeDateString(const CString& format);

static const int MAX_CUSTOM_COLORS = 16;

static const CString AllFilesFilter = TEXT("All Files|*.*|");

struct ConfigData {
	CString mDefExt;
	CString mTimeDateFormat;
	bool mRemoveTrailingWhitespaceOnSave;
	NewLineTypes mNewLineType;
	bool mAutoIndent;
	long mTabWidth;
	RecentFileList mRecentFileList;
	Commands mCommands;
	FindReplaceData mFindReplaceData;
	CRect mWinPos;
	bool mMaximized;
	DWORD mOpenFilterIndex;
	bool mShowLineNumbers;

	CFont mFont;
	ColorSchemes mColorSchemes;
	LanguageMgr mLangMgr;
	COLORREF mCustomColors[MAX_CUSTOM_COLORS];
	AccelTable mAccelTable;

	//this isn't persisted, but needs to be shared
	CString mTempClip;

	ConfigData();
	void DoPersist(bool load);
	void Load() {DoPersist(true);}
	void Save() {DoPersist(false);}
	void SetEdit(EditCtrl& edit);
	void SwitchClipboard(HWND hwnd);

	void Execute(const Command& cmd, const CString& fileName, HWND hwnd) const;
	void ExecuteHelp(const CString& topic, HWND hwnd) const;

private:
	void AddConfigDataForOldVersions(long version);
};

extern ConfigData gConfigData;

//The following isn't actually config data. It just needs to be shared
//across dialogs and isn't general enough to be in Common.

static const CString PROGRAM_NAME = TEXT("Lite Edit");
static const CString PROGRAM_NAME_AND_VER = PROGRAM_NAME + TEXT(" 1.1");

#endif