#include "StdAfx.h"
#include "EditCtrl.h"
#include <algorithm>
#include "Windowsx.h"

const UINT ID_SCROLLSELECT = 1;
const long BOOKMARK_WIDTH = 20;
const long SPACE_BETWEEN_GUTTER_AND_CLIENT_AREA = 1;//space between client area and where gutter appears to start
const COLORREF COLOR_BOOKMARK = RGB(0x0, 0xFF, 0xFF);
const COLORREF COLOR_LINENUMBER = RGB(0x0, 0x60, 0x0);
const COLORREF COLOR_GUTTER = RGB(230, 230, 230);

void TextHolder::UndoImpl(UndoStack& undos, UndoStack& redos, bool undo) {
	if ((undo && CanUndo()) || (!undo && CanRedo())) {
		m_undoOn = false;
		UndoItem item = undos.back();//make a copy so it can be used after erased from undos

		redos.push_back(item);
		undos.erase(&undos.back());

		const CString& add = undo ? item.deleted : item.added;
		const CString& del = undo ? item.added : item.deleted;

		//m_unmodifyUndoSize stores the size of m_undos when modified is set to false.
		//So we set modified to false iff m_undos.size() = m_unmodifyUndoSize.
		//See SetModified().
		bool modify = m_undos.size() != m_unmodifyUndoSize;

		ModifyText(item.pos, del.GetLength(), add, modify);

		m_undoOn = true;
	}
}

long TextHolder::ValidateLength() {
	long cutAmount = GetTextLength() - GetMaxLength();

	if (cutAmount > 0) {
		VERIFY(ModifyText(GetTextLength() - cutAmount, cutAmount, EmptyStr));
		return cutAmount;
	} else
		return 0;
}

void TextHolder::PushUndoItem(const UndoItem& item) {
	if (m_undoOn) {
		if (m_undoLimit == 0)
			m_undos.clear();
		else {
			//make sure m_undos doesn't exceed m_undoLimit
			if (m_undoLimit != -1 && m_undos.size() >= m_undoLimit)
				m_undos.erase(m_undos.begin(), m_undos.begin() + (m_undos.size() - m_undoLimit) + 1);
			
			m_undos.push_back(item);
			m_redos.clear();
		}
	}
}

//you can pass in the result of a prior call to GetPosInfo() as priorInfo to
//increase performance
PosInfo TextHolder::GetPosInfo(long pos, PosInfo priorInfo/* = PosInfo(0, 0)*/) const {
	ATLASSERT(0 <= pos && pos <= GetTextLength());
	ATLASSERT(priorInfo.pos <= pos);
	LPCTSTR start = GetText(), end = start + pos, p = start + priorInfo.pos;
	long lineNum = priorInfo.line;

	while (p < end)
		if (*p++ == '\n')
			++lineNum;

	return PosInfo(pos, lineNum);
}

bool TextHolder::ModifyText(long pos, long count, const CString& _newText, bool modified/* = true*/) {
	ATLASSERT(0 <= pos && pos <= GetTextLength());
	ATLASSERT(count >= 0);

	if (count <= 0 && (_newText.IsEmpty() || GetTextLength() == GetMaxLength())) {
		if (!_newText.IsEmpty())
			OnLimitText();

		return false;
	}

	//make sure newText won't make text longer than MaxLength()
	long newLen = GetTextLength() + _newText.GetLength() - count;
	CString newText = newLen > GetMaxLength() ? _newText.Left(GetMaxLength() - GetTextLength() + count) : _newText;

	OnPreChange(pos, count, newText);

	PushUndoItem(UndoItem(pos, m_text.Mid(pos, count), newText));

	long first = 0, last = 0;
	PosInfo start = GetPosInfo(pos);

	if (count > 0) {
		PosInfo end = GetPosInfo(pos + count, start);

		m_text.Delete(pos, count);

		m_lines[start.line]->Invalidate();
		if (end.line > start.line)
			m_lines.erase(&m_lines[start.line + 1], &m_lines[end.line+1]);

		first = last = start.line;
	}
	
	if (!newText.IsEmpty()) {
		m_text.Insert(pos, newText);
		
		PosInfo end = GetPosInfo(pos + newText.GetLength(), start);
		
		m_lines[start.line]->Invalidate();
		if (end.line > start.line)
			for (long i = start.line + 1; i <= end.line; ++i)
				m_lines.insert(&m_lines[i], NewLineData());

		first = start.line;
		last = end.line;
	}

	SetModified(modified);
	OnPostChange(pos, first, last, newText.GetLength(), count);
	return true;
}

void TextHolder::DeleteChar(long pos) {
	ATLASSERT(0 <= pos && pos < GetTextLength());

	OnPreChange(pos, 1, EmptyStr);

	PosInfo start = GetPosInfo(pos);
	PosInfo end = GetPosInfo(pos + 1, start);

	PushUndoItem(UndoItem(pos, m_text.GetAt(pos), NULL));

	m_text.Delete(pos, 1);

	m_lines[start.line]->Invalidate();
	if (start.line != end.line)
		m_lines.erase(&m_lines[end.line]);

	SetModified(true);
	OnPostChange(pos, start.line, start.line, 0, 1);
}

bool TextHolder::InsertChar(long pos, TCHAR ch) {
	ATLASSERT(0 <= pos && pos <= GetTextLength());

	if (GetTextLength() < GetMaxLength()) {
		CString strCh = CharToStr(ch);
		OnPreChange(pos, 0, strCh);

		PushUndoItem(UndoItem(pos, EmptyStr, strCh));

		m_text.Insert(pos, ch);

		PosInfo start = GetPosInfo(pos);
		PosInfo end = GetPosInfo(pos + 1, start);

		m_lines[start.line]->Invalidate();
		if (start.line != end.line)
			m_lines.insert(&m_lines[end.line], NewLineData());

		SetModified(true);
		OnPostChange(pos, start.line, end.line, 1, 0);
		return true;
	} else {
		OnLimitText();
		return false;
	}
}

void TextHolder::SetMaxLength(long newMaxLength) {
	if (newMaxLength < 0)
		newMaxLength = 0;

	m_maxLength = newMaxLength;
	ValidateLength();
}

//m_undos will not be modified, even if newLimit is < m_undoLimit
void TextHolder::SetUndoLimit(long newLimit) {
	ATLASSERT(newLimit >= -1);
	m_undoLimit = newLimit;
}

void TextHolder::SetModified(bool val) {
	m_modified = val;

	if (!m_modified && m_undoOn)
		m_unmodifyUndoSize = m_undos.size();
}

void EditCtrl::ValidatePos(long& pos, long maxLen/* = -1*/) const {
	pos = Min(pos, maxLen == -1 ? GetTextLength() : maxLen);
	pos = Max(pos, 0L);
}

int EditCtrl::GetCharWidth(long pos) const {
	int width = 1;
	long line, offset;
	CurDC dc(*this);
	LineFromChar(offset, line, pos);
	if (pos - offset < GetLineLength(line)) {
		TCHAR ch = GetText().GetAt(pos);
		//I can't use GetTextExtent() to calculate width of tab char since tab can
		//vary in width
		if (ch == '\t')
			width =
				dc.GetLineWidth(line, pos - offset + 1) -
				dc.GetLineWidth(line, pos - offset);
		else
			width = dc.GetTextExtent(CString(ch)).cx;
	} else
		width = GetAvgCharWidth();

	return width;
}

int EditCtrl::GetCaretWidth() const {
	return m_insert ? 2 : GetCharWidth(GetSelCaret());
}

void EditCtrl::CreateCaret() {
	if (HasFocus()) {
		VERIFY(CreateSolidCaret(GetCaretWidth(), GetCharHeight()));
		ShowCaret();
	}
}

long EditCtrl::GetLineWidth(long i, CurDC* dc/* = NULL*/) const {
	LineDataPtr lineData = GetLineData(i);

	if (lineData->m_width == -1)
		lineData->m_width = dc == NULL ?
			CurDC(*this).GetLineWidth(i, GetLineLength(i)) :
			dc->GetLineWidth(i, GetLineLength(i));

	return m_lines[i]->m_width;
}

void EditCtrl::ValidateWidestLine(long first, long last) {
	CurDC dc(*this);

	for (long i = first; i <= last; ++i)
		if (GetLineWidth(i) > GetLineWidth(m_widestLineIndex, &dc))
			m_widestLineIndex = i;
}

long EditCtrl::GetWidestLineWidth() const {
	return GetLineWidth(m_widestLineIndex);
}

//this is used in the debug build to make sure m_widestLineIndex is
//validated correctly
long EditCtrl::GetWidestLineWidthInefficiently() const {
	long width = 0;
	for (long i = 0; i < GetLineCount(); ++i)
		width = Max(width, GetLineWidth(i));

	return width;
}

const CodeLinePtr& EditCtrl::GetCodeLine(long i) const {
	LineDataPtr lineData = GetLineData(i);

	if (!lineData->m_codeLineValid) {
		lineData->m_codeLineValid = true;
		const Context& context = i == 0 ? Context() : GetCodeLine(i - 1)->GetEndingContext();
		lineData->m_codeLine = m_lang->NewCodeLine(GetLine(i), context);
	}

	return m_lines[i]->m_codeLine;
}

bool EditCtrl::IndentBlock(TCHAR indentChar, bool indent) {
	long selStart, selEnd;
	GetSelRange(selStart, selEnd);
	long startPos, endPos, startLine, endLine;
	LineFromChar(startPos, startLine, selStart);
	LineFromChar(endPos, endLine, selEnd);
	if (startLine == endLine)
		return false;

	//if any part of the line is selected, the line will be
	//indented, so select the whole thing
	if (selEnd > endPos)
		endPos += GetLineLength(endLine);

	--endPos;//don't get the final new line

	CString block = GetTextRange(startPos, endPos);

	bool modified = true;
	CString indentStr = CString('\n') + indentChar;
	if (indent) {
		block.Replace(TEXT("\n"), indentStr);
		block.Insert(0, indentChar);
	} else {
		modified = block.Replace(indentStr, TEXT("\n")) > 0;
		if (block.GetAt(0) == indentChar) {
			block.Delete(0);
			modified = true;
		}
	}

	if (modified) {
		ModifyText(startPos, endPos - startPos, block);
		SetSel(startPos, startPos + block.GetLength() + 1);
	}

	return true;
}

void EditCtrl::InvalidateTextRange(long start, long end) {
	if (start == end)
		return;

	long first, last, startOffset, endOffset;

	LineFromChar(startOffset, first, start);
	LineFromChar(endOffset, last, end);

	//only bother to invalidate the visible lines
	long firstVisible = Max(first, GetFirstVisibleLine());
	long lastVisible = Min(last, GetLastVisibleLine(false));

	CRect rc(0, (firstVisible - GetFirstVisibleLine()) * GetLineHeight(), 0, 0);
	rc.bottom = rc.top + GetLineHeight();
	long fvp = GetFirstVisiblePixel();

	CurDC dc(*this);

	for (long i = firstVisible; i <= lastVisible; ++i) {
		if (i == first && i == last) {
			rc.left = dc.GetLineWidth(i, start - startOffset) - fvp;
			rc.right = dc.GetLineWidth(i, end - endOffset) - fvp;
		} else if (i == first) {
			rc.left = dc.GetLineWidth(i, start - startOffset) - fvp;
			rc.right = GetLineWidth(i) - fvp;
		} else if (i == last) {
			rc.left = 0;
			rc.right = dc.GetLineWidth(i, end - endOffset) - fvp;
		} else {
			rc.left = 0;
			rc.right = GetLineWidth(i) - fvp;
		}

		//???
		//for some reason, if I didn't add 1 to the right of the rect of a partially
		//selected line, occasionally, not everything would be invalidated in a
		//non-monospace font
		//++rc.right;

		InvalidateRect(rc, TRUE);
		rc.OffsetRect(0, GetLineHeight());
	}
}

void EditCtrl::InvalidateSel() {
	long start, end;
	GetSelRange(start, end);
	InvalidateTextRange(start, end);
}

void EditCtrl::GetLineRange(long line, long& start, long& end) const {
	start = LineIndex(line);
	
	if (start == -1)
		start = end = 0;
	else {
		end = GetText().Find('\n', start);
		if (end == -1)
			end = GetTextLength();
	}
}

void EditCtrl::GetTextMetrics(TEXTMETRIC& tm) const {
	CurDC dc(*this);
	MemClear(tm);
	VERIFY(dc.GetTextMetrics(&tm));
}

long EditCtrl::GetAvgCharWidth() const {
	if (m_avgCharWidth == -1) {
		TEXTMETRIC tm;
		GetTextMetrics(tm);
		m_avgCharWidth = tm.tmAveCharWidth;
	}

	return m_avgCharWidth;
}

long EditCtrl::GetCharHeight() const {
	TEXTMETRIC tm;
	GetTextMetrics(tm);
	return tm.tmHeight;
}

void EditCtrl::ScrollCaret(ScrollKinds sk) {
	if (sk != skNone) {
		long lineOffset, lineNum;
		LineFromChar(lineOffset, lineNum, GetSelCaret());

		//Scroll vertically. Scroll as little as possible to make sure the line with the caret
		//is completely visible.
		if (lineNum < GetFirstVisibleLine())
			SetFirstVisibleLine(lineNum);
		else {
			if (GetLineCount() <= GetMaxVisibleLineCount(true))//make first line the first visible line is all lines are visible
				SetFirstVisibleLine(0);
			else {
				long last = GetLastVisibleLine(true);
				if (lineNum > last)
					//make lineNum the last visible line
					SetFirstVisibleLine(lineNum + GetFirstVisibleLine() - last);
			}
		}

		if (sk == skSelection) {
			//Scroll horizontally. We must make the caret visible. Also, we should make as much
			//of the selection visible as possible. We should scroll as little as possible to
			//do this.
			long start, end;
			GetLineSel(lineNum, start, end);
			CurDC dc(*this);
			long startPixel = dc.GetLineWidth(lineNum, start);//left end of selection
			long endPixel = dc.GetLineWidth(lineNum, end);//right end of selection
			bool caretOnLeft = GetSelActiveEnd() <= GetSelAnchor();

			const long fvp = GetFirstVisiblePixel();//the left side of the screen
			const long lvp = fvp + GetClientWidth();//the right side of the screen
			long scrollAmount = 0;

			if (start == end) {//if no selection
				//make sure caret itself is visible
				endPixel += GetCaretWidth();

				if (startPixel < fvp)
					scrollAmount = startPixel - fvp;
				else if (endPixel > lvp)
					scrollAmount = endPixel - lvp;
			} else {
				if (caretOnLeft) {
					if (startPixel < fvp)
						scrollAmount = startPixel - fvp;//scroll left
					else if (endPixel > lvp) {
						long amount1 = startPixel - fvp;
						long amount2 = endPixel - lvp;
						scrollAmount = Min(amount1, amount2);//scroll right
					}
				} else {
					if (endPixel > lvp)
						scrollAmount = endPixel - lvp;//scroll right
					else if (startPixel < fvp) {
						long amount1 = lvp - endPixel;
						long amount2 = fvp - startPixel;
						scrollAmount = -Min(amount1, amount2);//scroll left
					}
				}
			}

			SetFirstVisiblePixel(fvp + scrollAmount);
		} else if (sk == skCaret) {
			CurDC dc(*this);
			long curPixel = dc.GetLineWidth(lineNum, GetSelCaret() - lineOffset);
			const long hScrollCaretAmount = 15 * GetHScrollAmount();
			while (true) {
				if (curPixel < GetFirstVisiblePixel())
					SetFirstVisiblePixel(GetFirstVisiblePixel() - hScrollCaretAmount);
				else if (curPixel >= GetFirstVisiblePixel() + GetClientWidth())
					SetFirstVisiblePixel(GetFirstVisiblePixel() + hScrollCaretAmount);
				else
					break;
			}
		}
	}
}

void EditCtrl::UpdateHScroll() {
	SCROLLINFO si;
	InitSized(si);
	si.fMask = SIF_RANGE|SIF_PAGE|SIF_POS|SIF_DISABLENOSCROLL;
	si.nMin = 0;
	si.nMax = Div(GetWidestLineWidth(), GetHScrollAmount(), rkUp) - 1;
	si.nPage = GetClientWidth() / GetHScrollAmount();
	si.nPos = Div(GetFirstVisiblePixel(), GetHScrollAmount(), rkUp);
	SetScrollInfo(SB_HORZ, &si);
}

void EditCtrl::UpdateVScroll() {
	SCROLLINFO si;
	InitSized(si);
	si.fMask = SIF_RANGE|SIF_PAGE|SIF_POS|SIF_DISABLENOSCROLL;
	si.nMin = 0;
	si.nMax = GetLineCount() - 1;
	si.nPage = GetMaxVisibleLineCount(true);
	si.nPos = GetFirstVisibleLine();
	SetScrollInfo(SB_VERT, &si);
}

//Returns line relative indices of selection if part of line is selected. Returns 0 for
//start and end if no part of the line is selected.
void EditCtrl::GetLineSel(long line, long& start, long& end) const {
	long selStart, selEnd;
	GetSelRange(selStart, selEnd);

	long lineStart, lineEnd;
	GetLineRange(line, lineStart, lineEnd);

	if (selEnd < lineStart || selStart > lineEnd) {
		start = end = 0;
	} else {
		start = Max(selStart, lineStart) - lineStart;
		end = Min(selEnd, lineEnd) - lineStart;
	}
}

void EditCtrl::SelectWord(long pos) {
	long start, end;
	GetWordAt(pos, start, end);
	SetSel(start, end);
}

CString EditCtrl::GetPreSpace(long line, long caretOffset) const {
	ATLASSERT(0 <= line && line < GetLineCount());
	const CString& string = GetLine(line);

	LPCTSTR start = string, p = start;
	while (_istspace(*p))
		++p;

	long numChars = Min(long(p - start), caretOffset);
	return string.Left(numChars);
}

void EditCtrl::NotifyParent(WORD code) const {
	SendMessage(GetParent(), WM_COMMAND, MAKEWPARAM(GetDlgCtrlID(), code), (LPARAM)m_hWnd);
}

void EditCtrl::CalcGutterWidth() {
	long newGutterWidth = GetShowBookmarks() ? BOOKMARK_WIDTH : 0;

	if (GetShowLineNumbers()) {
		//make some extra space between line numbers and bookmark
		if (newGutterWidth > 0)
			newGutterWidth += 2;

		CurDC dc(*this);
		//I'm assuming no number is wider than 8
		newGutterWidth += GetNumDigits(GetLineCount()) * dc.GetTextExtent("8").cx;
	}
	
	if (newGutterWidth > 0)
		newGutterWidth += SPACE_BETWEEN_GUTTER_AND_CLIENT_AREA;

	if (m_gutterWidth != newGutterWidth) {
		m_gutterWidth = newGutterWidth;

		if (IsWindow()) {
			//I need to call SetWindowPos() with the SWP_FRAMECHANGED flag so
			//WM_NCCALCSIZE will be sent since the size of the non-client area
			//has changed. I've found a lot less flickering occurs if I also pass
			//SWP_NOREDRAW and invalidate the client and non-client area manually.
			SetWindowPos(NULL, CRect(), SWP_FRAMECHANGED|SWP_NOSIZE|SWP_NOMOVE|SWP_NOZORDER|SWP_NOREDRAW);
			Invalidate();
			SendMessage(WM_NCPAINT, 1);//redraw non-client area
		}
	}
}

CRect EditCtrl::GetGutterRect() const {
	CRect rc;
	rc.left = GetSystemMetrics(SM_CXEDGE);
	rc.top = GetSystemMetrics(SM_CYEDGE);
	rc.right = rc.left + m_gutterWidth;
	rc.bottom = rc.top + GetClientHeight();
	if (GetVScroll())
		rc.bottom += GetSystemMetrics(SM_CYHSCROLL);
	return rc;
}

void EditCtrl::InvalidateGutter() {
	CRgn rgn;
	VERIFY(rgn.CreateRectRgnIndirect(&GetGutterRect()));
	SendMessage(WM_NCPAINT, (WPARAM)(HRGN)rgn);
}

void EditCtrl::IncreaseSel(long amount) {
	long start, end;
	GetSel(start, end);
	SetSel(start, end + amount);//change the active end
}

void EditCtrl::OnPreChange(long pos, long count, const CString& newText) {
	m_preChangeLineCount = GetLineCount();

	m_visibleLines.clear();
	long first = GetFirstVisibleLine(), last = GetLastVisibleLine(false);
	for (long i = first; i <= last; ++i)
		m_visibleLines.push_back(GetCodeLine(i));

	//put selection at start of modification so that during modification
	//the text that the selection refers to doesn't change
	SetSelCaret(pos, false);
}

//pos is char index where text modification started. first and last are the first and last
//lines that were modified, respectively.
void EditCtrl::OnPostChange(long pos, long first, long last, long charsAdded, long charsDeleted) {
	long sizeChange = charsAdded - charsDeleted;
	long numLinesChange = GetLineCount() - m_preChangeLineCount;

	//make sure all code lines are valid because GetCodeLine() is recursive and a stack
	//overflow could otherwise result if too much text is inserted
	for (long i = first; i <= last; ++i)
		GetCodeLine(i);

	const long lastLine = GetLineCount() - 1;

	//validate the lineIndex's
	for (i = last + 1; i <= lastLine; ++i) {
		long& lineIndex = GetLineData(i)->m_lineIndex;
		if (lineIndex != -1)
			lineIndex += sizeChange;
	}

	//if modification caused inconsistency between the ending and starting contexts of
	//two consecutive lines, then lines should be invalidated until there are no more
	//inconsistencies
	i = last;
	while (i < lastLine) {
		const Context& endingContext = GetCodeLine(i)->GetEndingContext();
		const Context& startingContext = GetCodeLine(i+1)->GetStartingContext();
		if (endingContext == startingContext)
			break;

		++i;
		m_lines[i]->InvalidateCodeLine();
	}

	//make sure all cached code lines are valid to avoid possible stack overflow
	GetCodeLine(lastLine);

	//*** Validate m_widestLineIndex ***

	if (m_widestLineIndex < first)
		ValidateWidestLine(first, last);
	else {
		long oldLastLine = last - numLinesChange;
		//if widest line itself may have been modified, any line could
		//be the new longest line
		ATLASSERT(first <= m_widestLineIndex);
		if (m_widestLineIndex <= oldLastLine) {
			m_widestLineIndex = 0;
			ValidateWidestLine(1, GetLineCount() - 1);
		} else {
			//widest line is after the last modified line,
			//so adjust index to compensate for any added or deleted lines
			m_widestLineIndex += numLinesChange;
			ValidateWidestLine(first, last);
		}
	}

	//make sure the validation of m_widestLineIndex is correct
	ATLASSERT(GetWidestLineWidth() == GetWidestLineWidthInefficiently());

	//Special case to remove bookmark from first line if all text is deleted.
	//The bookmark is still removed if text was added after all the text was deleted.
	if (GetTextLength() == charsAdded && charsDeleted > 0 && GetLineData(0)->m_hasBookmark) {
		GetLineData(0)->m_hasBookmark = false;
		InvalidateGutter();
	}

	//figure out what needs to be invalidated due to change
	long firstVisibleLine = GetFirstVisibleLine();
	long lastVisibleLine = Max(GetLastVisibleLine(false), firstVisibleLine + (long)m_visibleLines.size() - 1L);
	const long lineCount = GetLineCount();
	CRect rc(0, 0, GetClientWidth(), GetLineHeight());
	
	for (i = firstVisibleLine; i <= lastVisibleLine; ++i) {
		//if indexing GetCodeLine() or m_visibleLines would not be valid,
		//invalidate the entire line
		if (i >= lineCount || i - firstVisibleLine >= m_visibleLines.size())
			InvalidateRect(rc);
		else {
			const CodeLine& oldLine = *m_visibleLines[i - firstVisibleLine];
			const CodeLine& newLine = *GetCodeLine(i);
			long diff = CodeLine::FindFirstDiff(oldLine, newLine);
			if (diff != -1) {
				CurDC dc(*this);
				rc.left = dc.GetLineWidth(i, diff) - GetFirstVisiblePixel();
				InvalidateRect(rc);
				rc.left = 0;
			}
		}
		
		rc.OffsetRect(0, GetLineHeight());
	}

	//validate first visible line
	SetFirstVisibleLine(GetFirstVisibleLine());

	//put selection at end of modification
	SetSelCaret(pos + charsAdded);

	//if the caret hasn't moved but chars were deleted in overwrite mode, then
	//the caret is on a different character
	if (charsAdded == 0 && charsDeleted > 0 && !GetInsert())
		CreateCaret();

	//If a line is added
	//or removed, a bookmarked may have been moved or delete, so the gutter may
	//need to be redrawn.
	//Also, if line numbers are being displayed, adding or deleting a line should
	//invalidate the gutter.
	if (numLinesChange != 0) {
		CalcGutterWidth();
		InvalidateGutter();
		UpdateVScroll();
	}

	UpdateHScroll();

	NotifyParent(ECN_CHANGE);
}

EditCtrl::EditCtrl() {
	m_font.Attach((HFONT)GetStockObject(DEFAULT_GUI_FONT));
	MemClear(m_sel);
	m_firstVisibleLine = 0;
	m_firstVisiblePixel = 0;
	m_avgCharWidth = m_lineHeight = -1;
	m_widestLineIndex = 0;
	m_preChangeLineCount = 0;
	m_insert = true;
	m_readOnly = false;
	m_autoIndent = true;
	m_greyedWhenReadOnly = false;
	m_useSystemSelectionColors = true;
	m_showBookmarks = true;
	m_showLineNumbers = false;
	CalcGutterWidth();
	SetTabWidth(4);
	m_curOffset = -1;
	m_mouseDownFromSelectWord = false;
}

bool EditCtrl::Create(HWND parent, RECT& rc, DWORD style, DWORD exStyle, UINT id/* = 0*/, LanguagePtr lang/* = NULL*/, TokenColorsPtr tokenColors/* = NULL*/) {
	SetLanguage(lang);
	SetTokenColors(tokenColors);
	m_hScroll = (style & WS_HSCROLL) != 0;
	m_vScroll = (style & WS_VSCROLL) != 0;
	bool result = WndBase::Create(parent, rc, NULL, style, exStyle, id) != NULL;
	SetHScroll(m_hScroll);
	SetVScroll(m_vScroll);
	return result;
}

void EditCtrl::SetText(const CString& newText) {
	//SetText() should clear all bookmarks.
	//OnPostChange() clear all bookmarks except the one on the first
	//line of the text is already empty empty, so I make a special case
	//for that here so all bookmarks get removed.
	if (GetTextLength() == 0 && GetLineData(0)->m_hasBookmark)
		ClearAllBookmarks();

	ModifyText(0, GetTextLength(), newText);
	//ModifyText() puts the caret at the end of the modification
	SetSelCaret(0);
}

CString EditCtrl::GetTextEx(NewLineTypes type) const {
	CString text = GetText();

	if (type == nlCR)
		ConvertLFtoCR(text);
	else if (type == nlCRLF)
		text = ConvertLFtoCRLF(text, GetLineCount());

	return text;
}

void EditCtrl::ReplaceSel(const CString& newSel, bool selectReplacment/* = false*/) {
	long start, end;
	GetSelRange(start, end);
	if (ModifyText(start, end - start, newSel))
		if (selectReplacment)
			SetSel(start, start + newSel.GetLength(), skSelection);
}

const CString& EditCtrl::GetLine(long i) const {
	LineDataPtr lineData = GetLineData(i);

	if (!lineData->m_lineValid) {
		lineData->m_lineValid = true;
		long start, end;
		GetLineRange(i, start, end);
		lineData->m_line = GetTextRange(start, end);
	}

	return lineData->m_line;
}

long EditCtrl::GetLineLength(long line) const {
	ATLASSERT(0 <= line && line < GetLineCount());

	LPCTSTR start = (LPCTSTR)GetText() + LineIndex(line), end = start;
	while (*end != NULL && *end != '\n')
		++end;

	return end - start;
}

long EditCtrl::LineIndex(long line) const {
	LineDataPtr lineData = GetLineData(line);
	long& lineIndex = lineData->m_lineIndex;

	if (lineIndex == -1) {
		if (line == 0)
			lineIndex = 0;
		else
			lineIndex = LineIndex(line-1) + GetLineLength(line-1) + 1;//+1 for new line
	}

	return lineIndex;

	/*
	ATLASSERT(0 <= line && line < GetLineCount());

	//this code is hard to read due to lexical tricks, but it NEEDS to be efficient
	LPCTSTR start = GetText(), p = start;
	while (line-- > 0)
		while (*p++ != '\n');

	return p - start;
	*/
}

//uses a binary search to find that line with a line index corresponding to lineIndex
long EditCtrl::FindLineIndex(long lineIndex, long first, long last) const {
	ATLASSERT(first <= last);//lineIndex is not an actual line index or my line index caches are wrong?

	long mid = (first + last) / 2;
	long midLineIndex = LineIndex(mid);

	if (midLineIndex == lineIndex)
		return mid;
	else if (midLineIndex < lineIndex)
		first = mid + 1;
	else
		last = mid - 1;

	return FindLineIndex(lineIndex, first, last);
}

void EditCtrl::LineFromChar(long& lineIndex, long& lineNum, long pos) const {
	ATLASSERT(0 <= pos && pos <= GetTextLength());

	LPCTSTR start = GetText(), p = start + pos;

	//calculate the line index
	while (p > start && *(p - 1) != '\n')
		--p;
	lineIndex = p - start;

	//use the line index to calculate lineNum
	lineNum = FindLineIndex(lineIndex, 0, GetLineCount() - 1);
}

void EditCtrl::SetFirstVisibleLine(long fvl) {
	fvl = Max(0L, fvl);
	long maxVisible = Min(GetMaxVisibleLineCount(true), GetLineCount());
	fvl = Min(GetLineCount() - maxVisible, fvl);

	if (m_firstVisibleLine != fvl) {
		int dy = (m_firstVisibleLine - fvl) * GetLineHeight();
		m_firstVisibleLine = fvl;
		VERIFY(ScrollWindow(0, dy));
		InvalidateGutter();

		//I can't update the scroll position when pressing up or down
		//because SetScrolPos() would slow scrolling down noticably.
		//The scroll position will get updated in OnKeyUp().
		if (!IsKeyDown(VK_UP) && !IsKeyDown(VK_DOWN))
			SetScrollPos(SB_VERT, fvl);
	}
}

long EditCtrl::GetLastVisibleLine(bool completelyVisible) const {
	return
		Min(GetFirstVisibleLine() + GetMaxVisibleLineCount(completelyVisible),
			GetLineCount()) - 1L;
}

//If completelyVisible is true, the number of lines that are entirely visible
//are returned.
//If completelyVisible is true, the number of lines that any part of are visible
//are returned.
long EditCtrl::GetMaxVisibleLineCount(bool completelyVisible) const {
	long lineHeight = GetLineHeight();
	return Div(GetClientHeight(), lineHeight, completelyVisible ? rkDown : rkUp);
}

void EditCtrl::SetFirstVisiblePixel(long fvp) {
	fvp = Min(fvp, GetWidestLineWidth() - GetClientWidth() + 1L);
	fvp = Max(0L, fvp);

	if (m_firstVisiblePixel != fvp) {
		int dx = m_firstVisiblePixel - fvp;
		m_firstVisiblePixel = fvp;
		VERIFY(ScrollWindow(dx, 0));

		//I can't update the scroll position when pressing up or down
		//because SetScrolPos() would slow scrolling down noticably.
		//The scroll position will get updated in OnKeyUp().
		if (!IsKeyDown(VK_LEFT) && !IsKeyDown(VK_RIGHT))
			SetScrollPos(SB_HORZ, Div(fvp, GetHScrollAmount(), rkUp));
	}
}

void EditCtrl::SetSel(long anchor, long end, ScrollKinds sk/* = skCaret*/) {
	ValidatePos(anchor);
	ValidatePos(end);

	bool selChanged = m_sel.cpMin != anchor || m_sel.cpMax != end;
	if (selChanged) {
		long oldStart, oldEnd;
		GetSelRange(oldStart, oldEnd);
		long newStart = Min(anchor, end);
		long newEnd = Max(anchor, end);

		m_sel.cpMin = anchor;
		m_sel.cpMax = end;

		//must call ScrollCaret() before PosFromChar() or InvalidateTextRange() because
		//ScrollCaret() modifies the first visible pixel and line which
		//PosFromChar() and InvalidateTextRange() use
		ScrollCaret(sk);

		SetCaretPos(PosFromChar(end));

		//recreate the caret when in overwrite mode
		if (end != oldEnd && !m_insert)
			CreateCaret();

		//I want to invalidate only the stuff that needs it, i.e. text that is selected that
		//was not before and text that is not selected that was before. I put the sections
		//of text over which the selection is changing into vec. Then I iterate through vec,
		//and any section in vec that is only inside the new selection or only inside the old
		//selection gets invalidated.
		vector<long> vec;
		vec.push_back(newStart);
		vec.push_back(newEnd);
		vec.push_back(oldStart);
		vec.push_back(oldEnd);
		sort(vec.begin(), vec.end());

		for (size_t i = 0; i < vec.size() - 1; ++i) {
			long start = vec[i], end = vec[i + 1];
			bool insideNew = start >= newStart && end <= newEnd;
			bool insideOld = start >= oldStart && end <= oldEnd;
			if (insideNew != insideOld)
				InvalidateTextRange(start, end);
		}

		NotifyParent(ECN_SELCHANGE);
	} else
		ScrollCaret(sk);
}

void EditCtrl::GetSelRange(long& start, long& end) const {
	start = Min(m_sel.cpMin, m_sel.cpMax);
	end = Max(m_sel.cpMin, m_sel.cpMax);
}

CString EditCtrl::GetSelText() const {
	long start, end;
	GetSelRange(start, end);
	return GetTextRange(start, end);
}

long EditCtrl::GetCaretLine() const {
	long line, offset;
	LineFromChar(offset, line, GetSelCaret());
	return line;
}

void EditCtrl::GotoLine(long line) {
	long index = LineIndex(line);
	SetSelCaret(index);
}

void EditCtrl::RecolorAll() {
	for (long i = 0; i < GetLineCount(); ++i) {
		GetLineData(i)->InvalidateCodeLine();
		//need to validate code line since GetCodeLine() is recursive
		//and I don't want a stack overflow
		GetCodeLine(i);
	}

	if (IsWindow()) {
		Invalidate();
		InvalidateGutter();//some of the background color is drawn in the gutter
	}
}

void EditCtrl::GetWordAt(long pos, long& start, long& end) {
	end = ::GetNextWordPos(GetText(), pos, false);
	start = ::GetPrevWordPos(GetText(), end, false);
}

CPoint EditCtrl::PosFromChar(long pos) const {
	long offset, num;
	LineFromChar(offset, num, pos);

	CurDC dc(*this);
	int cx = dc.GetLineWidth(num, pos - offset);
	return CPoint(cx - GetFirstVisiblePixel(), (num - GetFirstVisibleLine()) * GetLineHeight());
}

long EditCtrl::CharFromPos(const CPoint& pt) const {
	long line = pt.y / GetLineHeight() + GetFirstVisibleLine();
	if (line < 0)
		return 0;
	if (line >= GetLineCount())
		return GetTextLength();

	const CString& text = GetLine(line);
	int tabLength = GetTabWidthInPixels();

	HDC thisDc = ::GetDC(m_hWnd);
	HDC compatDc = CreateCompatibleDC(thisDc);
	ATLASSERT(compatDc != NULL);
	VERIFY(::ReleaseDC(m_hWnd, thisDc));
	CurDC dc(*this, compatDc);
	int minusFvp = -GetFirstVisiblePixel();
	int x = minusFvp;
	//draw the line one character at a time (on a memory DC) and return when
	//the width of the drawn characters exceeds pt.x
	for (long i = 0; i < text.GetLength(); ++i) {
		x += dc.TabbedTextOut(x, 0, (LPCTSTR)text + i, 1, 1, &tabLength, minusFvp).cx;
		if (pt.x < x)
			return i + LineIndex(line);
	}

	VERIFY(DeleteDC(compatDc));

	return LineIndex(line) + GetLineLength(line);
}

long EditCtrl::GetClientWidth() const {
	CRect rc;
	GetClientRect(rc);
	return rc.Width();
}

long EditCtrl::GetClientHeight() const {
	CRect rc;
	GetClientRect(rc);
	return rc.Height();
}

long EditCtrl::GetLineHeight() const {
	if (m_lineHeight == -1) {
		TEXTMETRIC tm;
		GetTextMetrics(tm);
		m_lineHeight = tm.tmHeight;
		ATLASSERT(m_lineHeight > 0);
	}

	return m_lineHeight;
}

void EditCtrl::SetFont(const CFont& font) {
	if (!FontsEqual(m_font, font)) {
		if ((HFONT)m_font != NULL)
			VERIFY(m_font.DeleteObject());
		m_font.Attach(CloneFont(font));

		//recalculate line widths
		for (long i = 0; i < GetLineCount(); ++i)
			GetLineData(i)->m_width = -1;

		ValidateWidestLine(0, GetLineCount() - 1);

		m_lineHeight = -1;
		m_avgCharWidth = -1;
		CalcGutterWidth();//line numbers are drawn with the current font
		CreateCaret();
		SetFirstVisibleLine(GetFirstVisibleLine());//validate first visible line
		SetFirstVisiblePixel(GetFirstVisiblePixel());//validate first visible pixel
		UpdateHScroll();
		UpdateVScroll();
		Invalidate();
	}
}

void EditCtrl::SetHScroll(bool val) {
	m_hScroll = val;
	ShowScrollBar(SB_HORZ, val);
	UpdateHScroll();
}

void EditCtrl::SetVScroll(bool val) {
	m_vScroll = val;
	ShowScrollBar(SB_VERT, val);
	UpdateVScroll();
}

bool EditCtrl::CanCut() const {
	return IsWriteable() && CanCopy();
}

bool EditCtrl::CanCopy() const {
	return GetTextLength() > 0;
}

bool EditCtrl::CanPaste() const {
	return IsWriteable() && IsClipboardFormatAvailable(CF_TEXT);
}

void EditCtrl::Cut() {
	if (CanCut()) {
		Copy();

		if (HasSel())
			DeleteSel();
		else {
			//delete the current line

			long caretLine = GetCaretLine();
			long start, end;
			GetLineRange(caretLine, start, end);

			//we need to delete a new line character also if there is more than one
			//line
			if (GetLineCount() > 1)
				//delete previous new line character if we're on the last line and
				//there is no trailing new line character
				if (caretLine == GetLineCount() - 1)
					--start;
				else
					++end;

			ModifyText(start, end - start, EmptyStr);
		}
	}
}

void EditCtrl::Copy() const {
	if (CanCopy()) {
		//Copy the selection if there is one. Otherwise, copy the current line.
		CString textToCopy = HasSel() ? GetSelText() : GetLine(GetCaretLine()) + TEXT("\n");
		VERIFY(SetClipboardString(textToCopy, m_hWnd, true));
	}
			
}

void EditCtrl::Paste() {
	if (CanPaste()) {
		CString clipboard;
		if (GetClipboardString(clipboard, m_hWnd, true))
			ReplaceSel(clipboard);
	}
}

void EditCtrl::ToggleBookmark() {
	LineDataPtr data = GetLineData(GetCaretLine());
	data->m_hasBookmark = !data->m_hasBookmark;
	InvalidateGutter();
}

void EditCtrl::NextBookmark(bool next) {
	long line = GetCaretLine();
	long i;

	if (next) {
		for (i = line + 1; i < GetLineCount(); ++i)
			if (GetLineData(i)->m_hasBookmark) {
				GotoLine(i);
				return;
			}

		for (i = 0; i < line; ++i)
			if (GetLineData(i)->m_hasBookmark) {
				GotoLine(i);
				return;
			}
	} else {
		for (i = line - 1; i >= 0; --i)
			if (GetLineData(i)->m_hasBookmark) {
				GotoLine(i);
				return;
			}

		for (i = GetLineCount() - 1; i > line; --i)
			if (GetLineData(i)->m_hasBookmark) {
				GotoLine(i);
				return;
			}
	}
}

void EditCtrl::ClearAllBookmarks() {
	for (long i = 0; i < GetLineCount(); ++i)
		GetLineData(i)->m_hasBookmark = false;

	InvalidateGutter();
}

void EditCtrl::SetInsert(bool b) {
	if (m_insert != b) {
		m_insert = b;
		CreateCaret();
	}
}

void EditCtrl::SetReadOnly(bool b) {
	if (b != m_readOnly) {
		m_readOnly = b;
		Invalidate();
	}
}

COLORREF EditCtrl::GetSelForeColor() const {
	if (HasFocus()) {
		if (GetUseSystemSelectionColors())
			return GetSysColor(COLOR_HIGHLIGHTTEXT);
		else
			return GetColor(tkBackground);
	} else
		return GetSysColor(COLOR_WINDOW);
}

COLORREF EditCtrl::GetSelBackColor() const {
	if (HasFocus()) {
		if (GetUseSystemSelectionColors())
			return GetSysColor(COLOR_HIGHLIGHT);
		else
			return GetColor(tkPlainText);
	} else
		return GetSysColor(COLOR_BTNSHADOW);
}

void EditCtrl::SetTabWidth(long newTabWidth) {
	if (m_tabWidth != newTabWidth) {
		m_tabWidth = newTabWidth;

		if (IsWindow())
			Invalidate();
	}
}

void EditCtrl::SetShowBookmarks(bool val) {
	m_showBookmarks = val;
	CalcGutterWidth();
}

void EditCtrl::SetShowLineNumbers(bool val) {
	m_showLineNumbers = val;
	CalcGutterWidth();
}

void EditCtrl::SetShowGreyedWhenReadOnly(bool val) {
	if (val != m_greyedWhenReadOnly) {
		m_greyedWhenReadOnly = val;
		Invalidate();
	}
}

void EditCtrl::SetUseSystemSelectionColors(bool val) {
	if (val != m_useSystemSelectionColors) {
		m_useSystemSelectionColors = val;
		Invalidate();
	}
}

static LanguagePtr EmptyLanguage = new Language();

void EditCtrl::SetLanguage(LanguagePtr lang) {
	if (lang == NULL)
		lang = EmptyLanguage;

	if (m_lang != lang) {
		m_lang = lang;
		RecolorAll();
	}
}

static TokenColorsPtr DefaultColors = new TokenColors();

void EditCtrl::SetTokenColors(TokenColorsPtr clrs) {
	if (clrs == NULL)
		clrs = DefaultColors;

	if (m_tokenColors != clrs) {
		m_tokenColors = clrs;
		RecolorAll();
	}
}

void EditCtrl::SetContextMenu(HMENU menu) {
	if (m_contextMenu != NULL)
		VERIFY(m_contextMenu.DestroyMenu());

	VERIFY(m_contextMenu.Attach(menu));
}

LRESULT EditCtrl::OnKeyDown(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	BYTE key = wParam;
	KeyModifiers mods = GetKeyModifiers();

	if (key == VK_UP || key == VK_DOWN) {
		long line, offset, pos = GetSelCaret();
		LineFromChar(offset, line, pos);

		bool willMove = (mods & kmShift) != 0 ?
			key == VK_UP || key == VK_DOWN :
			(key == VK_UP && line > 0) || (key == VK_DOWN && line < GetLineCount() - 1);

		if (willMove) {
			long fvp = GetFirstVisiblePixel();
			if (m_curOffset < 0)
				m_curOffset = PosFromChar(pos).x + fvp;//??? remove + GetCharWidth(pos) / 2;

			if (key == VK_UP)
				--line;
			else
				++line;

			long curLineY = GetLineHeight() * (line - GetFirstVisibleLine());
			long newPos = CharFromPos(CPoint(m_curOffset - fvp, curLineY));

			if ((mods & kmShift) != 0)
				IncreaseSel(newPos - pos);
			else
				SetSelCaret(newPos);

			//Ctrl + Up/Down scrolls
			if ((mods & kmCtrl) != 0) {
				long fvl = GetFirstVisibleLine();
				if (key == VK_UP)
					--fvl;
				else
					++fvl;

				SetFirstVisibleLine(fvl);
			}
		} else if (mods == kmNone && HasSel()) {
			RemoveSel();
			m_curOffset = -1;
		}
	} else if (key != VK_ESCAPE)
		m_curOffset = -1;

	switch (key) {
	case VK_LEFT :
		if (mods == kmNone)
			SetSelCaret(GetSelCaret() - 1);
		else if (mods == kmShift)
			IncreaseSel(-1);
		else if (mods == kmCtrl)
			SetSelCaret(GetPrevWordPos());
		else if (mods == (kmCtrl|kmShift))
			IncreaseSel(GetPrevWordPos() - GetSelCaret());
		break;
	case VK_RIGHT :
		if (mods == kmNone)
			SetSelCaret(GetSelCaret() + 1);
		else if (mods == kmShift)
			IncreaseSel(1);
		else if (mods == kmCtrl)
			SetSelCaret(GetNextWordPos());
		else if (mods == (kmCtrl|kmShift))
			IncreaseSel(GetNextWordPos() - GetSelCaret());
		break;
	case VK_HOME : {
		if (mods == kmNone || mods == kmShift) {
			long lineOffset, lineNum;
			LineFromChar(lineOffset, lineNum, GetSelCaret());

			if (mods == kmNone)
				SetSelCaret(lineOffset);
			else
				IncreaseSel(lineOffset - GetSelCaret());
		} else if (mods == kmCtrl)
			SetSelCaret(0);
		else if (mods == (kmShift|kmCtrl))
			IncreaseSel(-GetSelCaret());
		break;
	}
	case VK_END : {
		if (mods == kmNone || mods == kmShift) {
			long lineOffset, lineNum;
			LineFromChar(lineOffset, lineNum, GetSelCaret());
			long len = GetLineLength(lineNum);

			if (mods == kmNone)
				SetSelCaret(lineOffset + len);
			else
				IncreaseSel(len - (GetSelCaret() - lineOffset));
		} else if (mods == kmCtrl)
			SetSelCaret(GetTextLength());
		else if (mods == (kmShift|kmCtrl))
			IncreaseSel(GetTextLength() - GetSelCaret());
		break;
	}
	case VK_PRIOR :
		if ((mods == kmNone || mods == kmShift) && GetCaretLine() > 0) {
			CPoint pt = PosFromChar(GetSelCaret());
			pt.y -= GetClientHeight();
			pt.x += GetAvgCharWidth() / 2;
			long pos = CharFromPos(pt);
			SetFirstVisibleLine(GetFirstVisibleLine() - GetVisibleLineCount(true));
			if (mods == kmNone)
				SetSelCaret(pos);
			else
				IncreaseSel(pos - GetSelCaret());
		}
		break;
	case VK_NEXT :
		if ((mods == kmNone || mods == kmShift) && GetCaretLine() < GetLineCount() - 1) {
			CPoint pt = PosFromChar(GetSelCaret());
			//long lastLinePixel = GetClientHeight() - GetClientHeight();
			pt.y += GetClientHeight();
			pt.x += (GetAvgCharWidth() + 1) / 2;
			long pos = CharFromPos(pt);
			SetFirstVisibleLine(GetFirstVisibleLine() + GetVisibleLineCount(true));
			if (mods == kmNone)
				SetSelCaret(pos);
			else
				IncreaseSel(pos - GetSelCaret());
		}
		break;
		case 'A' :
			if (mods == kmCtrl)
				SelectAll();
			break;
		case VK_INSERT :
			if (mods == kmNone)
				SetInsert(!GetInsert());
			break;
		case VK_ESCAPE :
			if (mods == kmNone)
				SetSelCaret(GetSelCaret());//deselect text
			break;
	}

	if ((key == 'C' && mods == kmCtrl) || (key == VK_INSERT && mods == kmCtrl))
		SendMessage(WM_COPY);

	if (GetReadOnly())
		return 0;

	if ((key == 'V' && mods == kmCtrl) || (key == VK_INSERT && mods == kmShift))
		SendMessage(WM_PASTE);
	else if ((key == 'X' && mods == kmCtrl) || (key == VK_DELETE && mods == kmShift))
		SendMessage(WM_CUT);

	if (key == 'Z' && mods == kmCtrl)
		Undo();
	else if (key == 'Y' && mods == kmCtrl)
		Redo();

	switch (key) {
	case VK_DELETE :
		if (mods == kmNone) {
			if (HasSel())
				DeleteSel();
			else if (GetSelCaret() < GetTextLength())
				DeleteChar(GetSelCaret());
		}
		break;
	}

	return 0;
}

LRESULT EditCtrl::OnKeyUp(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	VirtualKey key = (VirtualKey)wParam;

	//I have to update the scroll position here because updating the scroll
	//position is too slow to do on every key down.

	if (key == VK_UP || key == VK_DOWN)
		if (GetScrollPos(SB_VERT) != GetFirstVisibleLine())
			SetScrollPos(SB_VERT, GetFirstVisibleLine());

	if (key == VK_LEFT || key == VK_RIGHT)
		if (GetScrollPos(SB_HORZ) != GetFirstVisiblePixel())
			SetScrollPos(SB_HORZ, Div(GetFirstVisiblePixel(), GetHScrollAmount(), rkUp));

	return 0;
}

LRESULT EditCtrl::OnChar(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	if (GetReadOnly())
		return 0;

	TCHAR key = wParam;
	if (key == '\r')
		key = '\n';

	if ((GetKeyModifiers() & kmCtrl) == 0 || key == '\n')
		switch (key) {
		case '\b' :
			if (HasSel())
				DeleteSel();
			else {
				long pos = GetSelCaret() - 1;
				if (pos >= 0)
					DeleteChar(pos);
			}
			break;
		case VK_ESCAPE :
			break;
		default : {
			if (HasSel() && (key == '\t' || key == ' '))
				if (IndentBlock(key, (GetKeyModifiers() & kmShift) == 0))
					break;

			if (HasSel())
				DeleteSel();

			if (m_insert || key == '\n' || GetChar(GetSelCaret()) == '\n' || GetSelCaret() == GetTextLength()) {
				if (key == '\n' && GetAutoIndent()) {
					long line, offset;
					LineFromChar(offset, line, GetSelCaret());
					CString space = '\n';
					if (GetSelCaret() > offset)
						space += GetPreSpace(line, GetSelCaret() - offset);
					ModifyText(GetSelCaret(), 0, space);
				} else		
					InsertChar(GetSelCaret(), key);
			} else {
				CString keyStr = key;
				ModifyText(GetSelCaret(), 1, keyStr);
			}
		}
		}

	return 0;
}

LRESULT EditCtrl::OnPaint(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	HideCaret();

	PAINTSTRUCT ps;
	CurDC dc(*this, BeginPaint(&ps));
	const CRect rcPaint = ps.rcPaint;

	bool hasFocus = HasFocus();
	int tabLength = GetTabWidthInPixels();

	long first = GetFirstVisibleLine() + rcPaint.top / GetLineHeight();
	first = Min(first, GetLineCount() - 1L);
	long last = GetFirstVisibleLine() + (rcPaint.bottom - 1) / GetLineHeight();
	last = Min(last, GetLineCount() - 1L);

	int y = (first - GetFirstVisibleLine()) * GetLineHeight();
	dc.SetWindowOrg(GetFirstVisiblePixel(), 0);

	for (long i = first; i <= last; ++i) {
		long selStart = 0, selEnd = 0;
		GetLineSel(i, selStart, selEnd);

		if (selEnd - selStart < GetLineLength(i)) {//if at least part of the line isn't selected
			const CodeLine& codeLine = *GetCodeLine(i);
			int x = 0;
			for (long i = 0; i < codeLine.size(); ++i) {
				const Token& token = codeLine[i];
				dc.SetTextColor(GetColor(token.GetKind()));
				dc.SetBkColor(GetBackColor());
				
				const CString& text = token.GetText();
				x += dc.TabbedTextOut(x, y, text, 1, &tabLength, 0).cx;
			}
		}
		
		if (selEnd != selStart) {
			CString sel = GetLine(i).Mid(selStart, selEnd - selStart);
			int left = dc.GetLineWidth(i, selStart);
			int right = dc.GetLineWidth(i, selEnd);

			const COLORREF selColor = GetSelForeColor();
			const COLORREF selBackColor = GetSelBackColor();
			dc.SetTextColor(selColor);
			dc.SetBkColor(selBackColor);
			CRect rcSel(left, y, right, y + GetLineHeight());
			dc.FillSolidRect(rcSel, selBackColor);
			dc.TabbedTextOut(left, y, sel, 1, &tabLength, 0);
		}

		y += GetLineHeight();
	}

	EndPaint(&ps);
	ShowCaret();
	return 0;
}

LRESULT EditCtrl::OnGetText(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	long copyCount = Max((long)wParam - 1L, 0L);
	StrCpy((LPTSTR)lParam, GetText().Left(copyCount));
	return copyCount;
}

LRESULT EditCtrl::OnSetText(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	SetText((LPCTSTR)lParam);
	return 0;
}

LRESULT EditCtrl::OnSetFocus(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	CreateCaret();
	InvalidateSel();

	bHandled = FALSE;
	return 0;
}

LRESULT EditCtrl::OnKillFocus(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	DestroyCaret();
	InvalidateSel();

	bHandled = FALSE;
	return 0;
}

LRESULT EditCtrl::OnGetDlgCode(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	if (IsKeyDown(VK_ESCAPE) && !HasSel())//get escape iff there is a selection
		return 0;//return 0 to NOT eat key
	if (GetReadOnly())
		return DLGC_WANTARROWS|DLGC_WANTCHARS;
	return DLGC_WANTALLKEYS|DLGC_WANTCHARS;
}

LRESULT EditCtrl::OnLButtonDblClk(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	CPoint pt(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
	long pos = CharFromPos(pt);
	SelectWord(pos);
	m_mouseDownFromSelectWord = true;
	return 0;
}

LRESULT EditCtrl::OnLButtonDown(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	SetFocus();
	CPoint pt(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
	long pos = CharFromPos(pt);

	if (wParam == (MK_CONTROL|MK_LBUTTON)) {
		SelectWord(pos);
		m_mouseDownFromSelectWord = true;
	} else if (wParam == (MK_SHIFT|MK_LBUTTON))
		IncreaseSel(pos - GetSelCaret());
	else if (wParam == MK_LBUTTON)
		SetSelCaret(pos);

	VERIFY(SetTimer(ID_SCROLLSELECT, 100) == ID_SCROLLSELECT);
	return 0;
}

LRESULT EditCtrl::OnLButtonUp(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	m_mouseDownFromSelectWord = false;
	return 0;
}

LRESULT EditCtrl::OnMouseMove(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	if (IsKeyDown(VK_LBUTTON) && !m_mouseDownFromSelectWord) {
		CPoint pt(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
		IncreaseSel(CharFromPos(pt) - GetSelCaret());
	}

	return 0;
}

LRESULT EditCtrl::OnCut(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	Cut();
	return 0;
}

LRESULT EditCtrl::OnCopy(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	Copy();
	return 0;
}

LRESULT EditCtrl::OnPaste(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	Paste();
	return 0;
}

LRESULT EditCtrl::OnVScroll(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	WORD pos = HIWORD(wParam);
	WORD code = LOWORD(wParam);
	if (code == SB_THUMBPOSITION || code == SB_THUMBTRACK)
		SetFirstVisibleLine(pos);
	else if (code == SB_LINEUP)
		SetFirstVisibleLine(GetFirstVisibleLine() - 1);
	else if (code == SB_LINEDOWN)
		SetFirstVisibleLine(GetFirstVisibleLine() + 1);
	else if (code == SB_PAGEUP)
		SetFirstVisibleLine(GetFirstVisibleLine() - GetVisibleLineCount(true));
	else if (code == SB_PAGEDOWN)
		SetFirstVisibleLine(GetFirstVisibleLine() + GetVisibleLineCount(true));


	return 0;
}

LRESULT EditCtrl::OnHScroll(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	WORD pos = HIWORD(wParam);
	WORD code = LOWORD(wParam);
	if (code == SB_THUMBPOSITION || code == SB_THUMBTRACK)
		SetFirstVisiblePixel(pos * GetHScrollAmount());
	else if (code == SB_LINEUP)
		SetFirstVisiblePixel(GetFirstVisiblePixel() - GetHScrollAmount());
	else if (code == SB_LINEDOWN)
		SetFirstVisiblePixel(GetFirstVisiblePixel() + GetHScrollAmount());
	else if (code == SB_PAGEUP)
		SetFirstVisiblePixel(GetFirstVisiblePixel() - GetClientWidth());
	else if (code == SB_PAGEDOWN)
		SetFirstVisiblePixel(GetFirstVisiblePixel() + GetClientWidth());

	return 0;
}

LRESULT EditCtrl::OnMouseWheel(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	short delta = HIWORD(wParam);
	long deltaLines = Div(3 * delta, WHEEL_DELTA, rkOff);
	SetFirstVisibleLine(GetFirstVisibleLine() - deltaLines);
	return 0;
}

LRESULT EditCtrl::OnSetCursor(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	WORD code = LOWORD(lParam);

	if (code == HTCLIENT) {
		SetCursor(LoadCursor(NULL, IDC_IBEAM));
		return TRUE;
	}
	
	bHandled = FALSE;
	return 0;
}

LRESULT EditCtrl::OnEraseBkGnd(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	RECT rc;
	GetClientRect(&rc);
	HDC dc = (HDC)wParam;
	CBrush brush;
	brush.CreateSolidBrush(GetBackColor());
	FillRect(dc, &rc, brush);
	return 1;
}

LRESULT EditCtrl::OnEnable(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	Invalidate();
	bHandled = FALSE;
	return 0;
}

//overridden to make bookmark gutter part of the non-client area
LRESULT EditCtrl::OnNcCalcSize(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	LRESULT result = DefWindowProc();

	RECT& rc = wParam ? ((NCCALCSIZE_PARAMS*)lParam)->rgrc[0] : *((RECT*)lParam);
	rc.left = Min(rc.left + m_gutterWidth, rc.right);

	return result;
}

//overridden to draw bookmarks
LRESULT EditCtrl::OnNcPaint(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	LRESULT result = DefWindowProc();

	if (m_gutterWidth == 0)
		return result;

    CDC dc;
	dc.Attach(GetWindowDC());

	CFont* oldFont = dc.SelectObject(CFont::FromHandle(GetFont()));

	CPen pen;
	pen.CreatePen(PS_SOLID, 1, RGB(0x0, 0x0, 0x0));
	CPen* oldPen = dc.SelectObject(&pen);
	CBrush brush;
	brush.CreateSolidBrush(COLOR_BOOKMARK);
	CBrush* oldBrush = dc.SelectObject(&brush);

	dc.SetTextColor(COLOR_LINENUMBER);
	dc.SetBkMode(TRANSPARENT);

	CRect rc = GetGutterRect();
	CRect space = rc;
	space.left = space.right - SPACE_BETWEEN_GUTTER_AND_CLIENT_AREA;
	rc.right -= SPACE_BETWEEN_GUTTER_AND_CLIENT_AREA;
	dc.FillSolidRect(&space, GetBackColor());
	dc.FillSolidRect(&rc, COLOR_GUTTER);
	rc.bottom = rc.top + GetLineHeight();
	rc.DeflateRect(0, 1);

	CRect rcBookmark = rc;
	rcBookmark.left = rcBookmark.right - BOOKMARK_WIDTH;
	CPoint pt(5, 5);

	long last = GetLastVisibleLine(false);
	CString lineNum;
	for (long i = GetFirstVisibleLine(); i <= last; ++i) {
		if (m_showLineNumbers) {
			lineNum.Format("%d", i + 1);
			dc.DrawText(lineNum, rc, DT_LEFT);
		}

		if (GetLineData(i)->m_hasBookmark)
			VERIFY(dc.RoundRect(rcBookmark, pt));

		rc.OffsetRect(0, GetLineHeight());
		rcBookmark.OffsetRect(0, GetLineHeight());
	}

	dc.SelectObject(oldPen);
	dc.SelectObject(oldBrush);
	dc.SelectObject(oldFont);

    ReleaseDC(dc.Detach());

	return result;
}

LRESULT EditCtrl::OnContextMenu(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	CPoint pt(lParam);

	if (pt.x == -1 && pt.y == -1) {//if context menu was invoked with keyboard
		pt = PosFromChar(GetSelCaret());
		ClientToScreen(&pt);
		pt.y += GetLineHeight();
	} else if (!HasSel()) {
		//move caret to where the user right-clicked if there is no selection
		//to make cut/copy of the currnet line behave more intuitively
		CPoint clientPt = pt;
		VERIFY(ScreenToClient(&clientPt));
		SetSelCaret(CharFromPos(clientPt));
	}

	VERIFY(TrackPopupMenu(m_contextMenu, 0, pt.x, pt.y, 0, GetParent(), NULL));

	return 0;
}

LRESULT EditCtrl::OnSize(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	UpdateVScroll();
	UpdateHScroll();
	return 0;
}

LRESULT EditCtrl::OnTimer(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	UINT id = (UINT)wParam;
	if (id == ID_SCROLLSELECT)
		if (IsAsyncKeyDown(VK_LBUTTON)) {
			CPoint pt;
			GetCursorPos(&pt);
			ScreenToClient(&pt);
			CRect rc;
			GetClientRect(&rc);

			if (!rc.PtInRect(pt))
				IncreaseSel(CharFromPos(pt) - GetSelCaret());
		} else
			VERIFY(KillTimer(ID_SCROLLSELECT));

	return 0;
}

#if 0
LRESULT EditCtrl::On(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	return 0;
}

#endif