#include "StdAfx.h"
#include "Utility.h"
#include "Persist.h"
#include "ShlObj.h"

LPCTSTR EmptyStr = TEXT("");
LPCTSTR ErrorStr = TEXT("Error");

void StringVector::GetLineRange(long line, long& start, long &end) const {
	ASSERT(0 <= line && line < size());
	start = 0;
	for (long i = 0; i <= line; ++i)
		if (i < line)
			start += Item(i).GetLength();
		else
			end = start + Item(i).GetLength();
}

long StringVector::GetTotalLength() const {
	long result = 0;

	for (size_type i = 0; i < size(); ++i)
		result += Item(i).GetLength();

	return result;
}

void StringVector::LineFromChar(long& lineOffset, long& lineNum, long pos) const {
	ASSERT(0 <= pos && pos <= GetTotalLength());
	lineOffset = 0;
	lineNum = -1;

	for (size_type i = 0; i < size(); ++i) {
		long length = Item(i).GetLength();

		if (i < size() - 1) {
			if (pos < length) {
				lineNum = i;
				break;
			}
		} else {
			if (pos <= length) {
				lineNum = i;
				break;
			}
		}

		lineOffset += length;
		pos -= length;
	}
}

CString StringVector::GetString(const char* separator) const {
	CString result;
	for (size_t i = 0; i < size(); ++i)
		AppendString(result, Item(i), separator);
	return result;
}

long StringVector::FindString(const CString& string, bool matchCase) const {
	for (size_t i = 0; i < size(); ++i)
		if (StrCmpEx(Item(i), string, matchCase) == 0)
			return i;

	return -1;
}

void StringVector::RemoveAllOccurrencesOf(const CString& string, bool matchCase) {
	size_t i = 0;
	while (i < size())
		if (StrCmpEx(Item(i), string, matchCase) == 0)
			erase(begin() + i);
		else
			++i;
}

void StringVector::RemoveDuplicates(bool matchCase) {
	size_t i1 = 0;
	while (i1 < size()) {
		size_t i2 = i1 + 1;
		while (i2 < size())
			if (StrCmpEx(Item(i1), Item(i2), matchCase) == 0)
				erase(begin() + i2);
			else
				++i2;

		++i1;
	}
}

CString StringVector::Concat(const CString& separator) {
	CString result;

	for (size_t i = 0; i < size(); ++i) {
		result += (*this)[i];
		if (i < (int)size() - 1)
			result += separator;
	}

	return result;
}

bool IsAsyncKeyDown(VirtualKey vk) {
	if (vk == VK_LBUTTON && GetSystemMetrics(SM_SWAPBUTTON))
		vk = VK_RBUTTON;
	else if (vk == VK_RBUTTON && GetSystemMetrics(SM_SWAPBUTTON))
		vk = VK_LBUTTON;

	return (GetAsyncKeyState(vk) & 0x8000) != 0;
}

KeyModifiers GetKeyModifiers() {
	KeyModifiers val = kmNone;

	if (IsKeyDown(VK_SHIFT))
		val |= kmShift;
	if (IsKeyDown(VK_CONTROL))
		val |= kmCtrl;
	if (IsKeyDown(VK_MENU))
		val |= kmAlt;

	return val;
}

CString KeyToString(VirtualKey vk) {
	if (vk == 0)
		return EmptyStr;

	LPARAM lParam = 0;
	lParam |= MapVirtualKey(vk, 0) << 16;
	lParam |= 1 << 25;//consider left and right of modifier keys the same

	CString result;
	const long SIZE = 20;
	LPSTR str = result.GetBuffer(SIZE);
	VERIFY(GetKeyNameText(lParam, str, SIZE) != 0);
	result.ReleaseBuffer();
	return result;
}

CString KeyAndModifiersToString(VirtualKey vk, KeyModifiers mods) {
	if (vk == 0)
		return EmptyStr;

	LPCTSTR separator = TEXT("+");
	CString result;

	if ((mods & kmCtrl) != 0)
		AppendString(result, KeyToString(VK_CONTROL), separator);
	if ((mods & kmShift) != 0)
		AppendString(result, KeyToString(VK_SHIFT), separator);
	if ((mods & kmAlt) != 0)
		AppendString(result, KeyToString(VK_MENU), separator);

	AppendString(result, KeyToString(vk), separator);
	return result;
}

WORD ToHotKeyMods(KeyModifiers mods) {
	WORD result = 0;
	if ((mods & kmAlt) != 0)
		result |= HOTKEYF_ALT;
	if ((mods & kmCtrl) != 0)
		result |= HOTKEYF_CONTROL;
	if ((mods & kmShift) != 0)
		result |= HOTKEYF_SHIFT;
	return result;
}

KeyModifiers FromHotKeyMods(WORD mods) {
	KeyModifiers result = kmNone;
	if ((mods & HOTKEYF_ALT) != 0)
		result |= kmAlt;
	if ((mods & HOTKEYF_CONTROL) != 0)
		result |= kmCtrl;
	if ((mods & HOTKEYF_SHIFT) != 0)
		result |= kmShift;
	return result;
}

void AppendString(CString& string, const CString& append, LPCTSTR separator) {
	if (!string.IsEmpty() && !append.IsEmpty())
		string += separator;
	string += append;
}

long GetNumCharsFit(CDC& dc, LPCTSTR string, long length, long maxWidth) {
	static SIZE szDummy;
	maxWidth = Max(0L, maxWidth);
	INT fit = 0;
	VERIFY(GetTextExtentExPoint(dc, string, length, maxWidth, &fit, NULL, &szDummy));

	for (long i = 0; i < fit; ++i)
		if (string[i] == '\n')
			return i + 1;

	return fit;
}

long GetPrevWordPos(const CString& string, long pos, bool eatSpaces) {
	ASSERT(0 <= pos && pos <= string.GetLength());

	if (eatSpaces)
		while (pos > 0 && _istspace(string.GetAt(pos-1)))
			--pos;

	if (pos > 0) {
		bool isVarChar = IsVarChar(string.GetAt(pos-1));
		while (
			pos > 0 &&
			isVarChar == IsVarChar(string.GetAt(pos-1)) &&
			!_istspace(string.GetAt(pos-1)))
			--pos;
	}

	return pos;
}

long GetNextWordPos(const CString& string, long pos, bool eatSpaces) {
	ASSERT(0 <= pos && pos <= string.GetLength());

	if (eatSpaces)
		while (pos < string.GetLength() && _istspace(string.GetAt(pos)))
			++pos;

	if (pos < string.GetLength()) {
		bool isVarChar = IsVarChar(string.GetAt(pos));
		while (
			pos < string.GetLength() &&
			isVarChar == IsVarChar(string.GetAt(pos)) &&
			!_istspace(string.GetAt(pos)))
			++pos;
	}

	if (eatSpaces)
		while (pos < string.GetLength() && _istspace(string.GetAt(pos)))
			++pos;

	return pos;
}

bool IsValidIdentifier(const CString& string) {
	ASSERT(!string.IsEmpty());
	if (string[0] != '_' && !_istalpha(string[0]))
		return false;

	for (int i = 1; i < string.GetLength(); ++i)
		if (!IsVarChar(string[i]))
			return false;

	return true;
}

long FindFirstDiff(LPCTSTR s1, LPCTSTR s2) {
	long result = 0;

	while (*s1 != NULL && *s2 != NULL) {
		if (*s1 != *s2)
			return result;

		++result;
		++s1;
		++s2;
	}

	return *s1 == NULL && *s2 == NULL ? -1 : result;
}

//this wrapper takes care of a const_cast on font and guarantees that lf is initialized
void GetLogFont(const CFont& font, LOGFONT& lf) {
	MemClear(lf);
	VERIFY(const_cast<CFont&>(font).GetLogFont(&lf));
}

HFONT CloneFont(const CFont& font) {
	if ((HFONT)font == NULL)//default font
		return NULL;

	LOGFONT lf;
	GetLogFont(font, lf);
	return CreateFontIndirect(&lf);
}

bool FontsEqual(const CFont& font1, const CFont& font2) {
	if (font1 == font2)
		return true;

	if ((HFONT)font1 == NULL || (HFONT)font2 == NULL)
		return false;

	LOGFONT lf1, lf2;
	GetLogFont(font1, lf1);
	GetLogFont(font2, lf2);

	return memcmp(&lf1, &lf2, sizeof(LOGFONT)) == 0;
}

long CountChar(const CString& string, TCHAR ch) {
	long result = 0;

	for (long i = 0; i < string.GetLength(); ++i)
		if (string.GetAt(i) == ch)
			++result;

	return result;
}

void MakeTitleCase(CString& string) {
	bool wordStart = true;
	for (int i = 0; i < string.GetLength(); ++i) {
		TCHAR ch = string.GetAt(i);
		string.SetAt(i, wordStart ? _totupper(ch) : _totlower(ch));
		wordStart = !_istalpha(ch);
	}
}

void MakeToggleCase(CString& string) {
	for (int i = 0; i < string.GetLength(); ++i) {
		TCHAR ch = string.GetAt(i);
		if (_istalpha(ch)) {
			ch = _istupper(ch) ? _totlower(ch) : _totupper(ch);
			string.SetAt(i, ch);
		}
	}
}

CString DupChar(TCHAR ch, long length) {
	CString result;
	LPTSTR buff = result.GetBuffer(length);
	for (long i = 0; i < length; ++i)
		buff[i] = ch;
	result.ReleaseBuffer(length);
	return result;
}

void StringListToStringVector(const CString& list, LPCTSTR separator, StringVector& vec) {
	int sepLen = _tcsclen(separator);
	ASSERT(sepLen > 0);

	vec.clear();
	int start = 0;
	while (start < list.GetLength()) {
		int end = list.Find(separator, start);
		if (end < 0)
			end = list.GetLength();

		CString item = list.Mid(start, end - start);
		item.TrimLeft(' ');
		item.TrimRight(' ');
		vec.push_back(item);
		start = end + sepLen;
	}
}

void UnescapeChars(CString& string) {
	CString s2;

	for (int i = 0; i < string.GetLength(); ++i) {
		TCHAR ch = string.GetAt(i);
		TCHAR nextCh = i + 1 < string.GetLength() ? string.GetAt(i+1) : NULL;
		if (ch == '\\' && (nextCh == '\\' || nextCh == 'n' || nextCh == 't')) {
			switch (nextCh) {
			case '\\' :
				s2 += '\\';
				break;
			case 'n' :
				s2 += '\n';
				break;
			case 't' :
				s2 += '\t';
				break;
			default :
				ASSERT(false);
			}
			++i;
		} else
			s2 += ch;
	}

	string = s2;
}

CString GetErrorMessage(DWORD errorCode) {
	const DWORD size = 0x400;
	TCHAR buff[size] = TEXT("");
	VERIFY(FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, errorCode, 0, buff, size, NULL) != 0);
	return buff;
}

TCHAR GetMnemonic(const CString& string) {
	CString noAmpersandString = string;
	//remove all occurrences of && which represent a single & in a menu
	//so I can find real mnemonics
	noAmpersandString.Replace(TEXT("&&"), EmptyStr);

	int pos = noAmpersandString.Find('&');
	if (0 <= pos && pos < noAmpersandString.GetLength() - 1)
		return _totupper(noAmpersandString.GetAt(pos + 1));
	else
		return NULL;
}

long Div(long num1, long num2, RoundKinds rk) {
	ASSERT(num2 != 0);

	switch (rk) {
	case rkUp :
		return (num1 + num2 - 1) / num2;
	case rkOff :
		return (num1 + num2 / 2) / num2;
	default :
		ASSERT(rk == rkDown);
		return num1 / num2;
	}
}

HMENU LoadSubMenu(int menuId, int pos) {
	CMenu menu;
	menu.Attach(LoadMenu(NULL, MAKEINTRESOURCE(menuId)));
	CMenu* subMenu = menu.GetSubMenu(pos);
	menu.RemoveMenu(pos, MF_BYPOSITION);//so subMenu won't be destroyed with menu
	return *subMenu;
}

void SetMenuItemEnabled(HMENU hMenu, UINT itemId, bool enable, bool byPos/* = false*/) {
	VERIFY(EnableMenuItem(hMenu, itemId, (byPos ? MF_BYPOSITION : MF_BYCOMMAND) | (enable ? MF_ENABLED : MF_DISABLED|MF_GRAYED)) != -1);
}

void SetSubMenu(HMENU menu, int subPos, HMENU subMenu) {
	CMenu m;
	VERIFY(m.Attach(menu));
	CString name;
	VERIFY(m.GetMenuString(subPos, name, MF_BYPOSITION));
	//ModifyMenu() always changes the menu text which is why I pass in name
	VERIFY(m.ModifyMenu(subPos, MF_BYPOSITION|MF_POPUP, (UINT)subMenu, name));
	m.Detach();
}

int FindMenuItemPosById(HMENU hMenu, UINT id) {
	CMenu menu;
	VERIFY(menu.Attach(hMenu));

	for (int i = 0; i < menu.GetMenuItemCount(); ++i)
		if (menu.GetMenuItemID(i) == id) {
			menu.Detach();
			return i;
		}

	menu.Detach();
	return -1;
}

void CleanupMenuItemName(CString& name) {
	ReplaceSubstring(name, TEXT("&"), EmptyStr, ffIgnoreCase);
	ReplaceSubstring(name, TEXT("..."), EmptyStr, ffIgnoreCase);
}

bool GetClipboardString(CString& text, HWND hwnd, bool convertFromCRLF) {
	bool success = false;

	if (OpenClipboard(hwnd)) {
		HANDLE handle = GetClipboardData(CF_TEXT);
		void* data = GlobalLock(handle);

		if (data != NULL) {
			text = (LPCTSTR)data;
			if (convertFromCRLF)
				ConvertCRLFtoLF(text);
			GlobalUnlock(handle);
			success = true;
		}

		BOOL closeClip = CloseClipboard();
		success = closeClip && success;
	}

	return success;
}

bool SetClipboardString(const CString& text, HWND hwnd, bool convertToCRLF) {
	bool success = false;

	if (OpenClipboard(hwnd)) {
		long extra = 1;//1 for NULL
		if (convertToCRLF)
			extra += CountChar(text, '\n');
		HGLOBAL hGlobal = GlobalAlloc(GMEM_MOVEABLE, text.GetLength() + extra);

		LPTSTR buff = (LPTSTR)GlobalLock(hGlobal);
		if (buff != NULL) {
			if (convertToCRLF)
				ConvertLFtoCRLF(buff, text);
			else
				strcpy(buff, text);

			if (EmptyClipboard()) {
				SetClipboardData(CF_TEXT, hGlobal);
				success = true;
			}

			GlobalUnlock(hGlobal);
		}

		BOOL closeClip = CloseClipboard();
		success = closeClip && success;
	}

	return success;
}

CString GetSystemTimeString(bool longTime) {
	SYSTEMTIME time;
	GetLocalTime(&time);

	DWORD flags = longTime ? 0 : TIME_NOSECONDS;
	int size = GetTimeFormat(LOCALE_USER_DEFAULT, flags, &time, NULL, NULL, 0);
	ASSERT(size != 0);
	CString result;
	TCHAR* buff = result.GetBuffer(size);
	size = GetTimeFormat(LOCALE_USER_DEFAULT, flags, &time, NULL, buff, size);
	ASSERT(size != 0);

	result.ReleaseBuffer(size - 1);//-1 to not include NULL
	return result;
}

CString GetSystemDateString(bool longDate) {
	SYSTEMTIME time;
	GetLocalTime(&time);

	DWORD flags = longDate ? DATE_LONGDATE : DATE_SHORTDATE;
	int size = GetDateFormat(LOCALE_USER_DEFAULT, flags, &time, NULL, NULL, 0);
	ASSERT(size != 0);
	CString result;
	TCHAR* buff = result.GetBuffer(size);
	size = GetDateFormat(LOCALE_USER_DEFAULT, flags, &time, NULL, buff, size);
	ASSERT(size != 0);

	result.ReleaseBuffer(size - 1);//-1 to not include NULL
	return result;
}

void ConvertCRLFtoLF(CString& string) {
	LPTSTR buff = string.GetBuffer(string.GetLength());
	ConvertCRLFtoLF(buff, buff);
	string.ReleaseBuffer();
}

//src uses \r\n for new lines. dset will use \n.
void ConvertCRLFtoLF(LPTSTR dest, LPCTSTR src) {
	while (true) {
		if (*src == '\r')//skip this char
			++src;

		ASSERT(*src != '\r');//had two consecutive \r's? Not valid CRLF string.
		*dest = *src;
		if (*dest == NULL)
			break;

		++dest;
		++src;
	}
}

CString ConvertLFtoCRLF(const CString& string, long lineCount/* = -1*/) {
	long lenNeeded = string.GetLength();
	lenNeeded += lineCount < 0 ? CountChar(string, '\n') : lineCount - 1;
	CString result;
	LPTSTR buff = result.GetBuffer(lenNeeded);
	ConvertLFtoCRLF(buff, string);
	result.ReleaseBuffer(lenNeeded);
	return result;
}

//src uses \n for new lines. dest will use \r\n.
void ConvertLFtoCRLF(LPTSTR dest, LPCTSTR src) {
	//not possible to convert if dest == src, unless src is empty
	ASSERT(dest != src || *src == NULL);

	while (true) {
		ASSERT(*src != '\r');//is src already CRLF?

		if (*src == '\n') {
			*dest = '\r';
			++dest;
		}

		*dest = *src;

		if (*dest == NULL)
			break;

		++dest;
		++src;
	}
}

void ConvertLFtoCR(CString& string) {
	long length = string.GetLength();
	LPTSTR buff = string.GetBuffer(length);
	ConvertLFtoCR(buff, buff);
	string.ReleaseBuffer(length);
}

void ConvertLFtoCR(LPTSTR dest, LPCTSTR src) {
	while (true) {
		*dest = *src == '\n' ? '\r' : *src;

		if (*dest == NULL)
			break;

		++dest;
		++src;
	}
}

void ConvertAnyToLF(CString& string) {
	LPTSTR buff = string.GetBuffer(string.GetLength());
	ConvertAnyToLF(buff, buff);
	string.ReleaseBuffer();
}

//converts a LF, CR, or CRLF string to a LF string
void ConvertAnyToLF(LPTSTR dest, LPCTSTR src) {
	while (true) {
		if (*src == '\r' && src[1] != '\n')//found just CR
			*dest = '\n';
		else {
			if (*src == '\r')//found CRLF, so skip the CR
				++src;

			ASSERT(*src != '\r');//should never happen

			*dest = *src;

			if (*dest == NULL)
				break;
		}

		++dest;
		++src;
	}
}

CString FileToString(const CString& fileName, bool convertNewLinesToLFs/* = true*/) {
	CFile file(fileName, CFile::modeRead);
	DWORD len = file.GetLength();
	CString result;
	LPSTR str = result.GetBuffer(len + 1);
	len = file.Read(str, len);
	str[len] = NULL;//Read() doesn't NULL terminate

	if (convertNewLinesToLFs)
		ConvertAnyToLF(str, str);

	result.ReleaseBuffer();
	return result;
}

void StringToFile(const CString& fileName, const CString& data) {
	UINT flags = CFile::modeWrite|CFile::modeCreate;
	CFile file(fileName, flags);
	file.Write((LPCTSTR)data, data.GetLength());
}

bool FileExists(const CString& name) {
	return GetFileAttributes(name) != -1;
}

bool AbsolutePathExists(const CString& path) {
	CString fullPath = path;
	return SearchForFile(fullPath) && fullPath.CompareNoCase(path) == 0;
}

//returns true if path is on this local machine, i.e. not a network path
bool IsLocalFile(const CString& path) {
	return path.GetLength() >= 2 && _istalpha(path[0]) && path[1] == ':';
}

//It looks like there's no good way to get the program files folder for
//all systems. CSIDL_PROGRAM_FILES is only defined for IE 5.0 and above.
//The ProgramFiles macro is not defined on all windows systems.
CString GetProgramFilesPath() {
	const int CSIDL_PROGRAM_FILES = 0x0026;
	TCHAR path[MAX_PATH];

	if (SHGetSpecialFolderPath(NULL, path, CSIDL_PROGRAM_FILES, TRUE))
		return path;
	else {
		const char* programFiles = getenv(TEXT("ProgramFiles"));

		if (programFiles != NULL)
			return programFiles;
		else
			return TEXT("C:\\Program Files");
	}
}

CString GetInternetExplorerPath() {
	return MakePath(GetProgramFilesPath(), TEXT("Internet Explorer\\iexplore.exe"));
}

CString GetModulePath() {
	CString modulePath;
	LPTSTR buff = modulePath.GetBuffer(MAX_PATH);
	DWORD len = GetModuleFileName(NULL, buff, MAX_PATH);
	ASSERT(len > 0);
	modulePath.ReleaseBuffer(len);
	return modulePath;
}

CString GetModuleDir() {
	return GetDir(GetModulePath());
}

CString MakePath(const CString& dir, const CString& fileName) {
	CString result = dir;

	if (!result.IsEmpty() && result.GetAt(result.GetLength()-1) != '\\')
		result += '\\';

	result += fileName;

	return result;
}

CString GetDir(const CString& path) {
	CString result = path;

	int pos = result.ReverseFind('\\');
	if (pos > 0) {
		++pos;//don't delete the trailing \
		result.Delete(pos, result.GetLength() - pos);
	}

	return result;
}

CString GetFileNameWithExt(const CString& path) {
	CString result = path;

	int pos = result.ReverseFind('\\');
	if (pos > 0)
		result.Delete(0, pos + 1);

	return result;
}

//returns true iff path has an extension
bool HasExt(const CString& path) {
	return !GetFileExt(path).IsEmpty();
}

//returns the extension of the file or path name, NOT including the dot.
//returns an empty string if the file name does not have an extension.
CString GetFileExt(const CString& path) {
	int findNameStart = Max(path.ReverseFind('\\'), 0);
	int pos = FindSubstring(path, TEXT("."), ffIgnoreCase|ffReverse, findNameStart);
	if (pos >= 0)
		return path.Right(path.GetLength() - pos - 1);
	else
		return EmptyStr;
}

bool IsValidFileName(const CString& fileName) {
	return fileName.FindOneOf(InvalidFileNameChars) < 0;
}

bool IsValidFileExt(const CString& ext) {
	return ext.FindOneOf(InvalidFileExtChars) < 0;
}

CString GetInvalidExtErrorMessage(const CString& ext) {
	ASSERT(!IsValidFileExt(ext));
	CString message;
	message.Format(TEXT("Extension '%s' is not valid. Extensions cannot contain any of the following characters:\n\t%s"), ext, InvalidFileExtChars);
	return message;
}

FILETIME GetFileModifyTime(const CString& fileName) {
	FILETIME fileTime;
	MemClear(fileTime);
	WIN32_FILE_ATTRIBUTE_DATA data;
	VERIFY(GetFileAttributesEx(fileName, GetFileExInfoStandard, &data));
	fileTime = data.ftLastWriteTime;
	return fileTime;
}

bool SearchForFile(CString& fileName) {
	DWORD len = SearchPath(NULL, fileName, NULL, 0, NULL, NULL);
	if (len == 0)//cannot find file
		return false;

	CString path;
	LPTSTR buff = path.GetBuffer(len);
	len = SearchPath(NULL, fileName, NULL, len, buff, NULL);
	ASSERT(len > 0);
	path.ReleaseBuffer(len);
	fileName = path;
	return true;
}

void BreakPath(const CString& path, CString& dir, CString& name, CString& ext) {
	int slash = path.ReverseFind('\\');
	int dot = path.ReverseFind('.');

	dir = slash >= 0 ? path.Left(slash + 1) : EmptyStr;
	ext = dot >= 0 ? path.Right((path.GetLength() - 1) - (dot + 1) + 1) : EmptyStr;

	int nameStart = slash >= 0 ? slash + 1 : 0;
	int nameEnd = dot >= 0 ? dot - 1 : path.GetLength() - 1;
	name = path.Mid(nameStart, nameEnd - nameStart + 1);
}

CString ExpandEnvironmentStrings(const CString& string) {
	DWORD size = ExpandEnvironmentStrings(string, NULL, 0);

	CString result;
	LPTSTR buff = result.GetBuffer(size);
	VERIFY(ExpandEnvironmentStrings(string, buff, size) != 0);
	result.ReleaseBuffer();
	return result;
}

bool BrowseForFolder(HWND hwnd, CString& path, UINT flags/* = 0*/) {
	LPTSTR buff = path.GetBufferSetLength(MAX_PATH);

	BROWSEINFO info;
	MemClear(info);
	info.hwndOwner = hwnd;
	info.pszDisplayName = buff;
	info.lpszTitle = TEXT("Choose installation directory:");
	info.ulFlags = flags;
	LPITEMIDLIST item = SHBrowseForFolder(&info);

	bool result = item != NULL;
	if (result) {
		result = SHGetPathFromIDList(item, buff) != FALSE;

		LPMALLOC malloc = NULL;
		HRESULT hr = SHGetMalloc(&malloc);

		if (SUCCEEDED(hr)) {
			malloc->Free(item);
			malloc->Release();
		} else
			ASSERT(false);
	}

	path.ReleaseBuffer();
	return result;
}

/******* This code was taken from some place in MSDN *******/

// CreateLink - uses the shell's IShellLink and IPersistFile interfaces
//   to create and store a shortcut to the specified object.
// Returns the result of calling the member functions of the interfaces.
// lpszPathObj - address of a buffer containing the path of the object.
// lpszPathLink - address of a buffer containing the path where the
//   shell link is to be stored.
// lpszDesc - address of a buffer containing the description of the
//   shell link.

HRESULT CreateLink(LPCSTR lpszPathObj, LPCSTR lpszPathLink, LPCSTR lpszDesc)
{
    HRESULT hres;
    IShellLink* psl;

    // Get a pointer to the IShellLink interface.
    hres = CoCreateInstance(CLSID_ShellLink, NULL,
        CLSCTX_INPROC_SERVER, IID_IShellLink, (LPVOID *) &psl);
    if (SUCCEEDED(hres)) {
        IPersistFile* ppf;

        // Set the path to the shortcut target and add the
        // description.
        VERSUCC(psl->SetPath(lpszPathObj));
        VERSUCC(psl->SetDescription(lpszDesc));
		VERSUCC(psl->SetWorkingDirectory(GetDir(lpszPathObj)));

       // Query IShellLink for the IPersistFile interface for saving the
       // shortcut in persistent storage.
        hres = psl->QueryInterface(IID_IPersistFile, (void**)&ppf);

        if (SUCCEEDED(hres)) {
            WCHAR wsz[MAX_PATH];

            // Ensure that the string is Unicode.
            MultiByteToWideChar(CP_ACP, 0, lpszPathLink, -1,
                wsz, MAX_PATH);

            // Save the link by calling IPersistFile::Save.
            hres = ppf->Save(wsz, TRUE);
            ppf->Release();
        }
        psl->Release();
    }
    return hres;
}

void GetSpecialFolderPath(int folder, TCHAR path[]) {
	//VERIFY(SUCCEEDED(SHGetFolderPath(NULL, folder, NULL, SHGFP_TYPE_CURRENT, path)));
	VERIFY(SHGetSpecialFolderPath(NULL, path, folder, TRUE));
}

//creates every directory in dir that does not exist.
//returns true if successful.
bool CreateDirPath(const CString& dir) {
	int pos = 0;
	CString parent;

	while (0 <= pos && pos < dir.GetLength()) {
		pos = dir.Find('\\', pos);
		if (pos < 0)
			parent = dir;
		else {
			parent = dir.Left(pos);
			++pos;
		}

		if (!FileExists(parent) && !CreateDirectory(parent, NULL))
			return false;
	}

	return true;
}

//This will only delete a directory if it is empty.
//It is NOT an error to try to delete a file that does not exist.
bool UninstallFile(const CString& path) {
	DWORD attrs = GetFileAttributes(path);
	if (attrs == -1)//if file does not exist
		return true;

	//remove read only so it can be deleted
	VERIFY(SetFileAttributes(path, attrs & ~FILE_ATTRIBUTE_READONLY));

	bool isDir = (attrs & FILE_ATTRIBUTE_DIRECTORY) != 0;
	BOOL res = isDir ? RemoveDirectory(path) : DeleteFile(path);

#if _DEBUG
	//this is useful for debugging
	if (!res) {
		DWORD err = GetLastError();
		CString msg;
		msg.Format(TEXT("Error while deleting file '%s'.\nError code: 0x%X."), path, err);
		MessageBox(NULL, msg, NULL, MB_OK);
	}
#endif

	return res != FALSE;
}

long GetNumDigits(long num) {
	ASSERT(num > 0);

	long digits = 0;
	while (num > 0) {
		++digits;
		num /= 10;
	}

	return digits;
}

void FindAllFiles(const CString& fileName, StringVector& vec, bool includeFullPath) {
	CFileFind find;
	vec.clear();

	BOOL found = find.FindFile(fileName);
	while (found) {
		found = find.FindNextFile();

		if (includeFullPath)
			vec.push_back(find.GetFilePath());
		else
			vec.push_back(find.GetFileName());
	}
}

void MoveWindowBy(CWnd& wnd, int dx, int dy) {
	CRect rc;
	wnd.GetWindowRect(&rc);
	wnd.GetParent()->ScreenToClient(&rc);
	rc.OffsetRect(dx, dy);
	wnd.MoveWindow(&rc);
}

bool FindItemByData(const CComboBox& comboBox, DWORD itemData, int& index) {
	for (index = 0; index < comboBox.GetCount(); ++index)
		if (comboBox.GetItemData(index) == itemData)
			return true;

	return false;
}

void SetListBoxItemText(CListBox& listBox, int index, const CString& text) {
	int sel = listBox.GetCurSel();
	VERIFY(listBox.DeleteString(index) != LB_ERR);
	VERIFY(listBox.InsertString(index, text) >= 0);
	listBox.SetCurSel(sel);
}

void StringVectorToComboBox(CComboBox& comboBox, const StringVector& vec) {
	comboBox.ResetContent();
	for (size_t i = 0; i < vec.size(); ++i)
		comboBox.AddString(vec[i]);
}

RECT GetScreenRect() {
	//??? this may not be multi monitor correct
	int cx = GetSystemMetrics(SM_CXFULLSCREEN);
	int cy = GetSystemMetrics(SM_CYFULLSCREEN);
	return CRect(0, 0, cx, cy);
}

void LinkifyText(CRichEditCtrl& richEdit, LPCTSTR text) {
	FINDTEXTEX ft;
	MemClear(ft);
	ft.chrg.cpMin = 0;
	ft.chrg.cpMax = richEdit.GetTextLength();
	ft.lpstrText = const_cast<LPTSTR>(text);

	long pos = richEdit.FindText(0, &ft);

	if (pos >= 0) {
		CHARRANGE cr;
		richEdit.GetSel(cr);

		CHARFORMAT2 cf;
		InitSized(cf);
		cf.dwMask = CFM_LINK;
		cf.dwEffects = CFE_LINK;
		richEdit.SetSel(ft.chrgText);
		VERIFY(richEdit.SetSelectionCharFormat(cf));

		richEdit.SetSel(cr);
	}
}

//Searches for an occurence of find inside search. end is actually one past the last
//index that you want to search. Pass -1 to end to specify the end of the string.
long FindSubstring(const CString& search, const CString& find, FindFlags flags, long start/* = 0*/, long end/* = -1*/) {
	if (end == -1)
		end = search.GetLength();

	ASSERT(0 <= start && (start < search.GetLength() || search.IsEmpty()));
	ASSERT(0 <= end && end <= search.GetLength());
	ASSERT(start <= end);
	ASSERT(!find.IsEmpty());

	bool matchCase = (flags & ffMatchCase) != 0;
	bool ignoreCase = (flags & ffIgnoreCase) != 0;
	bool wholeWord = (flags & ffWholeWord) != 0;
	bool reverse = (flags & ffReverse) != 0;

	//must pass in ffMatchCase or ffIgnoreCase, but not both
	ASSERT(matchCase != ignoreCase);

	for (long i1 = reverse ? end-1 : start; start <= i1 && i1 < end; reverse ? --i1 : ++i1) {
		//if wholeWord, then i1 must be the first char in the string or either
		//the i1th char or the (i1 - 1)th must be not be alphabetic
		if (wholeWord && i1 > 0) {
			bool isNotWholeWord =
				_istalpha(search.GetAt(i1 - 1)) &&
				_istalpha(search.GetAt(i1));

			if (isNotWholeWord)
				continue;
		}

		bool found = true;
		for (long i2 = 0; i2 < find.GetLength(); ++i2) {
			found =
				i1 + i2 < search.GetLength() &&
				ChrCmp(search.GetAt(i1 + i2), find.GetAt(i2), ignoreCase);

			if (!found)
				break;
		}

		//if wholeWord, then i1 must be the last char in the string or either the
		//i1th char or the (i1 + 1)th must not be alphabetic
		if (found && wholeWord) {
			long lastCharIndex = i1 + find.GetLength() - 1;
			found =
				lastCharIndex >= search.GetLength() - 1 ||
				!_istalpha(search.GetAt(lastCharIndex)) ||
				!_istalpha(search.GetAt(lastCharIndex+1));
		}

		if (found)
			return i1;
	}

	return -1;
}

//Searches for an occurence of find inside search. The search starts at start and ends
//at end. If end is less than or equal to start, the search will go from start to
//the end of the string, and then start again from the beginning of the string and go
//to end; hence, the search is cyclic. end is actually one past the last index that you
//want to search. You can NOT pass -1 for end to specify the length of search.
long CyclicFindSubstring(const CString& search, const CString& find, FindFlags flags, long start, long end) {
	ASSERT(end >= 0);//only FindSubstring() accepts -1 for the end of the string.

	if (start == search.GetLength())
		start = 0;

	if (end > start)
		return FindSubstring(search, find, flags, start, end);
	else {
		if ((flags & ffReverse) == 0) {//if forwards
			long pos = FindSubstring(search, find, flags, start, -1);
			if (pos < 0)
				pos = FindSubstring(search, find, flags, 0, end);
			return pos;
		} else {//if reverse
			long pos = FindSubstring(search, find, flags, 0, end);
			if (pos < 0)
				pos = FindSubstring(search, find, flags, start, -1);
			return pos;
		}
	}
}

long ReplaceSubstring(CString& string, const CString& find, const CString& replace, FindFlags flags, long start/* = 0*/, long maxReplacments/* = -1*/) {
	ASSERT(0 <= start && (start < string.GetLength() || string.IsEmpty()));

	long pos = start;
	long numReplacments = 0;

	while (true) {
		if (pos >= string.GetLength())
			break;
		if (numReplacments >= maxReplacments && maxReplacments >= 0)
			break;

		pos = FindSubstring(string, find, flags, pos);
		if (pos < 0)
			break;

		string.Delete(pos, find.GetLength());
		string.Insert(pos, replace);
		pos += replace.GetLength();
		++numReplacments;
	}

	return numReplacments;
}

long FindMatchImpl(const StringIterator& vec, const CString& string, bool matchCase, long first, long last) {
	if (first > last)
		return -1;

	long mid = (first + last) / 2;
	int comp = StrCmpEx(vec.GetNthString(mid), string, matchCase);

	if (comp == 0)
		return mid;
	else if (comp < 0)
		first = mid + 1;
	else
		last = mid - 1;

	return FindMatchImpl(vec, string, matchCase, first, last);
}

//Returns index of string into vec. Uses a binary search. vec must be sorted
//case sensetively if matchCase is true. vec must be sorted case insensetively
//if matchCase is false.
long FindMatch(const StringIterator& vec, const CString& string, bool matchCase) {
	return FindMatchImpl(vec, string, matchCase, 0, vec.GetNumStrings() - 1);
}

StringToken GetNextToken(const CString& string, long start, CString& token) {
	const long length = string.GetLength();
	ASSERT(0 <= start && start < length);

	TCHAR ch = string.GetAt(start);
	long pos = start + 1;
	StringToken result;

	if (_istspace(ch)) {
		while (pos < length && _istspace(string.GetAt(pos)))
			++pos;

		result = stWhiteSpace;
	} else if (_istalpha(ch) || ch == '_') {
		while (pos < length && IsVarChar(string.GetAt(pos)))
			++pos;

		result = stIdentifier;
	} else {
		while (pos < length) {
			ch = string.GetAt(pos);
			if (_istspace(ch) || ch == '_' || _istalpha(ch))
				break;

			++pos;
		}

		result = stSymbol;
	}

	token = string.Mid(start, pos - start);
	return result;
}

Accel::Accel() {
	key = 0;
	fVirt = 0;
	cmd = 0;
}

Accel::Accel(VirtualKey vkey, KeyModifiers mods, WORD command) {
	fVirt = FVIRTKEY;
	if ((mods & kmShift) != 0)
		fVirt |= FSHIFT;
	if ((mods & kmCtrl) != 0)
		fVirt |= FCONTROL;
	if ((mods & kmAlt) != 0)
		fVirt |= FALT;

	key = vkey;
	cmd = command;
}

KeyModifiers Accel::GetModifiers() const {
	KeyModifiers mods = 0;
	if ((fVirt & FSHIFT) != 0)
		mods |= kmShift;
	if ((fVirt & FCONTROL) != 0)
		mods |= kmCtrl;
	if ((fVirt & FALT) != 0)
		mods |= kmAlt;
	return mods;
}

void Accel::DoPersist(Persist& p) {
	p.PersistNumber(TEXT("Command"), cmd);
	p.PersistNumber(TEXT("Key"), key);
	p.PersistNumber(TEXT("Flags"), fVirt);
}

vector<Accel>::const_iterator AccelTable::FindAccelByCmd(UINT cmd) const {
	for (vector<Accel>::const_iterator it = m_accels.begin(); it != m_accels.end(); ++it)
		if (it->cmd == cmd)
			return it;

	return m_accels.end();
}

bool AccelTable::GetHotKey(UINT cmd, VirtualKey& key, KeyModifiers& mods) const {
	vector<Accel>::const_iterator it = FindAccelByCmd(cmd);
	if (it != m_accels.end()) {
		key = it->GetVirtualKey();
		mods = it->GetModifiers();
		return true;
	} else
		return false;
}

UINT AccelTable::FindHotKey(VirtualKey key, KeyModifiers mods) const {
	if (key != 0)
		for (size_t i = 0; i < m_accels.size(); ++i)
			if (m_accels[i].GetVirtualKey() == key && m_accels[i].GetModifiers() == mods)
				return m_accels[i].cmd;

	return 0;
}

void AccelTable::DecorateMenu(HMENU hMenu) {
	CMenu menu;
	menu.Attach(hMenu);

	for (size_t i = 0; i < m_accels.size(); ++i) {
		CString string;
		menu.GetMenuString(m_accels[i].cmd, string, MF_BYCOMMAND);

		//remove old shortcut
		long pos = string.Find('\t');
		if (pos != -1)
			string.Delete(pos, string.GetLength() - pos);

		CString hotkey = KeyAndModifiersToString(m_accels[i].GetVirtualKey(), m_accels[i].GetModifiers());
		AppendString(string, hotkey, TEXT("\t"));
		//ModifyMenu may fail if the command isn't in the menu which is OK.
		menu.ModifyMenu(m_accels[i].cmd, MF_BYCOMMAND|MF_STRING, m_accels[i].cmd, string);
	}

	menu.Detach();
}

void AccelTable::DoPersist(Persist& p) {
	p.PersistClassVector(TEXT("AcceleratorTable"), m_accels);

	if (p.IsLoading())
		Destroy();
}

UINT CommandIdCreator::CreateId() {
	if (mOldIds.empty())
		return mCurNewId++;
	else {
		UINT newId = mOldIds.back();
		mOldIds.erase(&mOldIds.back());
		return newId;
	}
}

void CommandIdCreator::ReleaseId(UINT id) {
	if (id >= mBase)//only release id if it was created by CreatorId
		mOldIds.push_back(id);
}

//safe to call if some sub menu items where not creator by CommandIdCreator
void CommandIdCreator::ReleaseSubMenuItemsIds(HMENU hMenu) {
	CMenu menu;
	VERIFY(menu.Attach(hMenu));

	for (int i = 0; i < menu.GetMenuItemCount(); ++i)
		ReleaseId(menu.GetMenuItemID(i));

	menu.Detach();
}

CString Exception::GetErrorMessage() const {
	CString msg;

	switch (mErrorCode) {
	case ErrNewerFileVersion :
		msg = TEXT("The file version is newer than the application version.");
		break;
	default : {
		ASSERT(false);
		msg.Format(TEXT("Undefined error %d"), mErrorCode);
	}
	}

	return msg;
}
